import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Literal, TypeAlias

from pybag.io.raw_reader import BaseReader, BytesReader
from pybag.io.raw_writer import CrcWriter
from pybag.mcap.chunk import decompress_chunk
from pybag.mcap.crc import assert_data_crc, assert_summary_crc
from pybag.mcap.error import McapNoChunkIndexError, McapNoSummarySectionError
from pybag.mcap.record_encoder import McapRecordWriter
from pybag.mcap.record_parser import (
    FOOTER_SIZE,
    MAGIC_BYTES_SIZE,
    McapRecordParser,
    McapRecordType
)
from pybag.mcap.records import (
    AttachmentIndexRecord,
    ChannelRecord,
    ChunkIndexRecord,
    DataEndRecord,
    FooterRecord,
    MessageIndexRecord,
    MessageRecord,
    MetadataIndexRecord,
    RecordType,
    SchemaRecord,
    StatisticsRecord,
    SummaryOffsetRecord
)
from pybag.types import Message

ChannelId: TypeAlias = int
"""Integer representing the channel ID."""

SchemaId: TypeAlias = int
"""Integer representing the schema ID."""

RecordId: TypeAlias = int
"""ID of an MCAP record (i.e. the record type)."""

LogTime: TypeAlias = int
"""Integer representing the log time."""

Offset: TypeAlias = int
"""Integer representing an offset from the start of a file/bytes."""

# Footer payload size: 8 bytes summary_start + 8 bytes summary_offset_start + 4 bytes summary_crc
FOOTER_PAYLOAD_SIZE = 20

# TODO: Summary should load enough to work without reading summary section again

class McapSummary(ABC):
    @abstractmethod
    def next_schema_id(self) -> SchemaId:
        ...  # pragma: no cover

    @abstractmethod
    def next_channel_id(self) -> ChannelId:
        ...  # pragma: no cover

    @abstractmethod
    def next_sequence_id(self, channel_id: ChannelId) -> int:
        ...  # pragma: no cover

    @abstractmethod
    def get_channel_id(self, topic: str) -> ChannelId | None:
        ...  # pragma: no cover

    @abstractmethod
    def get_schema_id(self, message: type[Message]) -> SchemaId | None:
        ...  # pragma: no cover

    @abstractmethod
    def get_schemas(self) -> dict[SchemaId, SchemaRecord]:
        ...  # pragma: no cover

    @abstractmethod
    def get_channels(self) -> dict[ChannelId, ChannelRecord]:
        ...  # pragma: no cover

    @abstractmethod
    def get_statistics(self) -> StatisticsRecord | None:
        ...  # pragma: no cover

    @abstractmethod
    def get_chunk_indexes(self) -> list[ChunkIndexRecord]:
        ...  # pragma: no cover

    @abstractmethod
    def get_message_indexes(
        self,
        chunk_index: ChunkIndexRecord,
    ) -> dict[ChannelId, MessageIndexRecord]:
        ...  # pragma: no cover

    @abstractmethod
    def get_attachment_indexes(self) -> dict[str, list[AttachmentIndexRecord]]:
        ...  # pragma: no cover

    @abstractmethod
    def get_metadata_indexes(self) -> dict[str, list[MetadataIndexRecord]]:
        ...  # pragma: no cover

    @abstractmethod
    def add_schema(self, schema: SchemaRecord):
        ...  # pragma: no cover

    @abstractmethod
    def add_channel(self, channel: ChannelRecord):
        ...  # pragma: no cover

    @abstractmethod
    def add_message(self, message: MessageRecord):
        ...  # pragma: no cover

    @abstractmethod
    def add_attachment_index(self, attachment_index: AttachmentIndexRecord):
        ...  # pragma: no cover

    @abstractmethod
    def add_metadata_index(self, metadata_index: MetadataIndexRecord):
        ...  # pragma: no cover

    @abstractmethod
    def write_summary(self, writer: CrcWriter):
        ...  # pragma: no cover


class McapChunkedSummary(McapSummary):
    """Summary information for a chunked MCAP file.

    This class handles loading summary information from an MCAP file, either from
    existing summary sections or by reconstructing it from the data section.

    Args:
        file: The file reader to read from.
        enable_crc_check: Whether to validate the CRC checksums
        enable_reconstruction: Controls reconstruction behavior:
            - 'never': Raise error if summary sections are missing
            - 'missing': Load from summary if present, otherwise reconstruct
            - 'always': Always reconstruct even if summary exists
    """

    def __init__(
        self,
        file: BaseReader | None = None,
        *,
        enable_crc_check: bool = False,
        enable_reconstruction: Literal['never', 'missing', 'always'] = 'missing',
        load_summary_eagerly: bool = True,
    ) -> None:
        logging.debug('Creating McapChunkedSummary')
        self._file = file
        self._enable_reconstruction = enable_reconstruction
        self._check_crc = enable_crc_check
        self._load_summary_eagerly = load_summary_eagerly

        # Initialize cache variables
        self._summary_offset: dict[RecordId, Offset] = {}
        self._cached_schemas: dict[SchemaId, SchemaRecord] = {}
        self._cached_channels: dict[ChannelId, ChannelRecord] = {}
        self._cached_chunk_indexes: list[ChunkIndexRecord] = []
        self._cached_metadata_indexes: dict[str, list[MetadataIndexRecord]] = defaultdict(list)
        self._cached_attachment_indexes: dict[str, list[AttachmentIndexRecord]] = defaultdict(list)
        self._message_indexes: dict[Offset, dict[ChannelId, MessageIndexRecord]] = {}
        # If self._file is None, then we start a new summary and so we can track stats from the beginning
        self._cached_statistics: StatisticsRecord | None = StatisticsRecord() if self._file is None else None

        # If we have no file, then we have the summary built (i.e. nothing)
        self._has_built_summary = self._file is None

        # Keep track of what we have searched for already
        self._has_searched_schemas = False
        self._has_searched_channels = False
        self._has_searched_chunks = False
        self._has_searched_attachments = False
        self._has_searched_metadata = False

        # If no file, then we are creating a new summary
        if self._file is None:
            logging.debug('Summary has no file to search')
            return

        # Read footer to determine if summary sections exist
        _ = self._file.seek_from_end(FOOTER_SIZE + MAGIC_BYTES_SIZE)
        self._footer = McapRecordParser.parse_footer(self._file)

        self._has_summary = self._footer.summary_start != 0
        self._has_summary_offset = self._footer.summary_offset_start != 0

        # Validate CRCs if requested
        if enable_crc_check:
            assert_data_crc(self._file, self._footer)
            assert_summary_crc(self._file, self._footer)

        if enable_reconstruction == 'never':
            if not self._has_summary:
                error_msg = 'No summary section detected in MCAP'
                raise McapNoSummarySectionError(error_msg)

        if enable_reconstruction == 'always' or not self._has_summary:
            logging.debug('Building summary from data section')
            self._build_summary()
            self._has_built_summary = True
        else:
            # TODO: What do we do if the summary is incomplete?
            logging.debug('Loading summary from summary section')
            self._load_summary()
            if self._cached_statistics is None and self._load_summary_eagerly:
                logging.warning('Statistics record not found in summary, generating')
                self._cached_statistics = self.get_statistics()

        # After loading summary, if we have records, then we assume they are complete
        # and we disable searching through the file for more records of the same type
        self._has_searched_schemas = len(self._cached_schemas) > 0
        self._has_searched_channels = len(self._cached_channels) > 0
        self._has_searched_chunks = len(self._cached_chunk_indexes) > 0
        self._has_searched_attachments = len(self._cached_attachment_indexes) > 0
        self._has_searched_metadata = len(self._cached_metadata_indexes) > 0

        # If not chunk index records are found, then we cannot use this class
        # TODO: For MCAP files that do not have chunks, this could be slow
        if not self.get_chunk_indexes() and self._file:
            raise McapNoChunkIndexError("No ChunkIndex records found!")

    def _load_summary(self) -> None:
        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Load summary: file not initialized!')
            return None

        # Seek to start of summary section
        _ = self._file.seek_from_start(self._footer.summary_start)

        # Iterate through records until we reach the footer
        while (record_type := McapRecordParser.peek_record(self._file)) != McapRecordType.FOOTER:
            # Exit when we reach the end of the summary section
            if record_type == McapRecordType.SUMMARY_OFFSET:
                break

            if record_type == McapRecordType.SCHEMA:
                if schema := McapRecordParser.parse_schema(self._file):
                    self._cached_schemas[schema.id] = schema
            elif record_type == McapRecordType.CHANNEL:
                channel = McapRecordParser.parse_channel(self._file)
                self._cached_channels[channel.id] = channel
            elif record_type == McapRecordType.ATTACHMENT_INDEX:
                attachment_index = McapRecordParser.parse_attachment_index(self._file)
                self._cached_attachment_indexes[attachment_index.name].append(attachment_index)
            elif record_type == McapRecordType.METADATA_INDEX:
                metadata_index = McapRecordParser.parse_metadata_index(self._file)
                self._cached_metadata_indexes[metadata_index.name].append(metadata_index)
            elif record_type == McapRecordType.STATISTICS:
                self._cached_statistics = McapRecordParser.parse_statistics(self._file)
            elif record_type == McapRecordType.CHUNK_INDEX:

                chunk_index = McapRecordParser.parse_chunk_index(self._file)
                self._cached_chunk_indexes.append(chunk_index)
            else:
                logging.error(f'Unexpected record in summary: {record_type}')
                McapRecordParser.skip_record(self._file)

        # Sort chunk indexes by message start time
        # TODO: Is this the best place to do this?
        self._cached_chunk_indexes.sort(key=lambda ci: ci.message_start_time)

    def _build_summary(self) -> None:
        """Build summary information by scanning the data section.

        This method populates instance variables directly and is only called
        when no summary section exists.
        """
        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Build summary: file not initialized!')
            return None

        # Track message statistics
        message_count = 0
        chunk_count = 0
        attachment_count = 0
        metadata_count = 0
        message_start_time: int | None = None
        message_end_time: int | None = None
        channel_message_counts: dict[ChannelId, int] = defaultdict(int)

        found_schemas: dict[SchemaId, SchemaRecord] = {}
        found_channels: dict[ChannelId, ChannelRecord] = {}
        found_chunk_indexes: list[ChunkIndexRecord] = []

        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)

        # Iterate through records until we reach the footer
        while (record_type := McapRecordParser.peek_record(self._file)) != McapRecordType.FOOTER:
            if record_type == McapRecordType.SCHEMA:
                if schema := McapRecordParser.parse_schema(self._file):
                    found_schemas[schema.id] = schema
            elif record_type == McapRecordType.CHANNEL:
                channel = McapRecordParser.parse_channel(self._file)
                found_channels[channel.id] = channel
            elif record_type == McapRecordType.CHUNK:
                chunk_count += 1

                chunk_start_offset = self._file.tell()
                chunk = McapRecordParser.parse_chunk(self._file)
                chunk_length = self._file.tell() - chunk_start_offset

                # Read the message index records
                message_index_length = 0
                reconstruct_message_index = True
                message_index_offsets: dict[int, int] = {}
                if self._enable_reconstruction != 'always':  # i.e. if 'missing'
                    start_post_chunk_message_index_offset = self._file.tell()
                    while McapRecordParser.peek_record(self._file) == McapRecordType.MESSAGE_INDEX:
                        post_chunk_message_index_offset = self._file.tell()
                        post_chunk_message_index = McapRecordParser.parse_message_index(self._file)
                        message_index_offsets[post_chunk_message_index.channel_id] = post_chunk_message_index_offset
                    message_index_length = self._file.tell() - start_post_chunk_message_index_offset
                    # If message index records exist, assume they are complete
                    reconstruct_message_index = not message_index_offsets
                    if reconstruct_message_index:
                        logging.warning("No message indexes found for chunk!")

                # Continue parsing the chunk
                compression = chunk.compression
                chunk_message_start_time: int | None = None
                chunk_message_end_time: int | None = None
                chunk_message_indexes: dict[ChannelId, list[tuple[LogTime, Offset]]] = defaultdict(list)

                compressed_size = len(chunk.records)
                reader = BytesReader(decompress_chunk(chunk, check_crc=self._check_crc))
                uncompressed_size = reader.size()

                while chunk_record_type := McapRecordParser.peek_record(reader):
                    if chunk_record_type == McapRecordType.SCHEMA:
                        if schema := McapRecordParser.parse_schema(reader):
                            found_schemas[schema.id] = schema
                    elif chunk_record_type == McapRecordType.CHANNEL:
                        channel = McapRecordParser.parse_channel(reader)
                        found_channels[channel.id] = channel
                    elif chunk_record_type == McapRecordType.MESSAGE:
                        message_offset = reader.tell()
                        chunk_message = McapRecordParser.parse_message(reader)
                        channel_id = chunk_message.channel_id
                        log_time = chunk_message.log_time

                        message_count += 1
                        channel_message_counts[channel_id] += 1
                        if reconstruct_message_index:
                            chunk_message_indexes[channel_id].append((log_time, message_offset))

                        if message_start_time is None or log_time < message_start_time:
                            message_start_time = log_time
                        if message_end_time is None or log_time > message_end_time:
                            message_end_time = log_time

                        if chunk_message_start_time is None or log_time < chunk_message_start_time:
                            chunk_message_start_time = log_time
                        if chunk_message_end_time is None or log_time > chunk_message_end_time:
                            chunk_message_end_time = log_time
                    elif chunk_record_type == McapRecordType.METADATA:
                        logging.warning(f'Metadata in chunk! This is not allowed in mcap spec. Ignoring...')
                        McapRecordParser.skip_record(reader)
                    elif chunk_record_type == McapRecordType.ATTACHMENT:
                        logging.warning(f'Attachment in chunk! This is not allowed in mcap spec. Ignoring...')
                        McapRecordParser.skip_record(reader)
                    else:
                        McapRecordParser.skip_record(reader)

                self._message_indexes[chunk_start_offset] = {
                    channel_id: MessageIndexRecord(
                        channel_id=channel_id,
                        records=sorted(records, key=lambda x: (x[0], x[1])),
                    )
                    for channel_id, records in chunk_message_indexes.items()
                }

                found_chunk_indexes.append(
                    ChunkIndexRecord(
                        message_start_time=chunk_message_start_time or 0,
                        message_end_time=chunk_message_end_time or 0,
                        chunk_start_offset=chunk_start_offset,
                        chunk_length=chunk_length,
                        message_index_offsets=message_index_offsets,
                        message_index_length=message_index_length,
                        compression=compression,
                        compressed_size=compressed_size,
                        uncompressed_size=uncompressed_size,
                    )
                )
            elif record_type == McapRecordType.MESSAGE:
                # TODO: Is there a way not to ignore?
                logging.warning('Found Message record outside of a chunk! Ignoring...')
                McapRecordParser.skip_record(self._file)
            elif record_type == McapRecordType.ATTACHMENT:
                attachment_count += 1
                start_offset = self._file.tell()
                attachment_record = McapRecordParser.parse_attachment(self._file)
                attachment_index = AttachmentIndexRecord(
                    offset=start_offset,
                    length=self._file.tell() - start_offset,
                    log_time=attachment_record.log_time,
                    create_time=attachment_record.create_time,
                    data_size=len(attachment_record.data),
                    name=attachment_record.name,
                    media_type=attachment_record.media_type,
                )
                self._cached_attachment_indexes[attachment_record.name].append(attachment_index)
            elif record_type == McapRecordType.METADATA:
                metadata_count += 1
                start_offset = self._file.tell()
                metadata_record = McapRecordParser.parse_metadata(self._file)
                metadata_index = MetadataIndexRecord(
                    offset=start_offset,
                    length=self._file.tell() - start_offset,
                    name=metadata_record.name,
                )
                self._cached_metadata_indexes[metadata_record.name].append(metadata_index)
            else:
                McapRecordParser.skip_record(self._file)

        self._cached_statistics = StatisticsRecord(
            message_count=message_count,
            schema_count=len(found_schemas),
            channel_count=len(found_channels),
            attachment_count=attachment_count,
            metadata_count=metadata_count,
            chunk_count=chunk_count,
            message_start_time=message_start_time or 0,
            message_end_time=message_end_time or 0,
            channel_message_counts=channel_message_counts,
        )
        self._cached_schemas = found_schemas
        self._cached_channels = found_channels
        self._cached_chunk_indexes = found_chunk_indexes

        # Sort chunk indexes by message start time
        # TODO: Is this the best place to do this?
        self._cached_chunk_indexes.sort(key=lambda ci: ci.message_start_time)

    def next_schema_id(self) -> int:
        # Schema ID must be non-zero
        return max(self.get_schemas().keys(), default=0) + 1

    def next_channel_id(self) -> int:
        # No harm in starting channel IDs from 1 too
        return max(self.get_channels().keys(), default=0) + 1

    def next_sequence_id(self, channel_id: ChannelId) -> int:
        if self._cached_statistics is not None:
            # We should always start at 1 because 0 indicates "not relevant"
            return self._cached_statistics.channel_message_counts.get(channel_id, 0) + 1
        return 0

    def get_channel_id(self, topic: str) -> ChannelId | None:
        for id, record in self._cached_channels.items():
            if record.topic == topic:
                return id
        return None

    def get_schema_id(self, message: type[Message]) -> SchemaId | None:
        for id, record in self._cached_schemas.items():
            if record.name == message.__msg_name__:
                return id
        return None

    def add_schema(self, schema: SchemaRecord):
        if schema.id in self._cached_schemas:
            logging.warning(f'Schema (id {schema.id}) already written to file')
        else:
            self._cached_schemas[schema.id] = schema
            if self._cached_statistics is not None:
                self._cached_statistics.schema_count += 1

    def add_channel(self, channel: ChannelRecord):
        if channel.id in self._cached_channels:
            logging.warning(f'Channel (id {channel.id}) already written to file')
        else:
            self._cached_channels[channel.id] = channel
            if self._cached_statistics is not None:
                self._cached_statistics.channel_count += 1
                self._cached_statistics.channel_message_counts[channel.id] = 0

    def add_message(self, message: MessageRecord):
        # Update statistics
        if self._cached_statistics is not None:
            self._cached_statistics.message_count += 1
            self._cached_statistics.channel_message_counts[message.channel_id] += 1
            self._cached_statistics.message_start_time = min(
                message.log_time if self._cached_statistics.message_start_time <= 0 else self._cached_statistics.message_start_time,
                message.log_time,
            )
            self._cached_statistics.message_end_time = max(
                message.log_time if self._cached_statistics.message_end_time <= 0 else self._cached_statistics.message_end_time,
                message.log_time,
            )

    def add_attachment_index(self, attachment_index: AttachmentIndexRecord):
        self._cached_attachment_indexes[attachment_index.name].append(attachment_index)
        if self._cached_statistics is not None:
            self._cached_statistics.attachment_count += 1

    def add_metadata_index(self, metadata_index: MetadataIndexRecord):
        self._cached_metadata_indexes[metadata_index.name].append(metadata_index)
        if self._cached_statistics is not None:
            self._cached_statistics.metadata_count += 1

    def add_chunk_index(self, chunk_index: ChunkIndexRecord, length: int):
        # Track chunk index for summary
        self._cached_chunk_indexes.append(chunk_index)
        if self._cached_statistics is not None:
            self._cached_statistics.chunk_count += 1

    def get_schemas(self) -> dict[SchemaId, SchemaRecord]:
        """Get all schemas defined in the MCAP file.

        Uses lazy loading: checks cache first, then loads from summary if available,
        otherwise returns pre-built data from _build_summary().

        Returns:
            Dictionary mapping schema ID to SchemaRecord
        """
        # If we have already searched through the data section, there is nothing else
        # we can do to find more schema records, so we return what we have.
        if self._has_built_summary or self._has_searched_schemas:
            return self._cached_schemas

        # If reconstruction not allowed, return empty
        if self._enable_reconstruction == 'never':
            logging.warning('No schema records found in summary and searching is disabled.')
            return self._cached_schemas

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Schemas: file not initialized!')
            return self._cached_schemas

        # Search for schema records in the summary section
        logging.warning('No schema records found in summary. Searching through file!')
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)
        while McapRecordParser.peek_record(self._file) != McapRecordType.DATA_END:
            next_record = McapRecordParser.peek_record(self._file)
            if next_record == McapRecordType.SCHEMA:
                if record := McapRecordParser.parse_schema(self._file):
                    self._cached_schemas[record.id] = record
            elif next_record == McapRecordType.CHUNK:
                chunk = McapRecordParser.parse_chunk(self._file)
                reader = BytesReader(decompress_chunk(chunk, check_crc=self._check_crc))
                while chunk_record_type := McapRecordParser.peek_record(reader):
                    if chunk_record_type == McapRecordType.SCHEMA:
                        if schema := McapRecordParser.parse_schema(reader):
                            self._cached_schemas[schema.id] = schema
                    else:
                        McapRecordParser.skip_record(reader)
            else:
                McapRecordParser.skip_record(self._file)
        self._has_searched_schemas = True
        return self._cached_schemas

    def get_channels(self) -> dict[ChannelId, ChannelRecord]:
        """Get all channels defined in the MCAP file.

        Uses lazy loading: checks cache first, then loads from summary if available,
        otherwise returns pre-built data from _build_summary().

        Returns:
            Dictionary mapping channel ID to ChannelRecord
        """
        # If we have already searched through the data section, there is nothing else
        # we can do to find more channel records, so we return what we have.
        if self._has_built_summary or self._has_searched_channels:
            return self._cached_channels

        # If reconstruction not allowed, return empty
        if self._enable_reconstruction == 'never':
            logging.warning('No channel records found in summary and reconstruction is disabled.')
            self._cached_channels = {}
            return self._cached_channels

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Channels: file not initialized!')
            return self._cached_channels

        # Search for channel records in the summary section
        logging.warning('No channel records found in summary. Searching through file!')
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)
        while McapRecordParser.peek_record(self._file) != McapRecordType.DATA_END:
            next_record = McapRecordParser.peek_record(self._file)
            if next_record == McapRecordType.CHANNEL:
                channel = McapRecordParser.parse_channel(self._file)
                self._cached_channels[channel.id] = channel
            elif next_record == McapRecordType.CHUNK:
                chunk = McapRecordParser.parse_chunk(self._file)
                reader = BytesReader(decompress_chunk(chunk, check_crc=self._check_crc))
                while chunk_record_type := McapRecordParser.peek_record(reader):
                    if chunk_record_type == McapRecordType.CHANNEL:
                        channel = McapRecordParser.parse_channel(reader)
                        self._cached_channels[channel.id] = channel
                    else:
                        McapRecordParser.skip_record(reader)
            else:
                McapRecordParser.skip_record(self._file)
        self._has_searched_channels = True
        return self._cached_channels

    def get_chunk_indexes(self) -> list[ChunkIndexRecord]:
        """Get all chunk indexes, sorted by message_start_time.

        Uses lazy loading: checks cache first, then loads from summary if available,
        otherwise returns pre-built data from _build_summary().

        Returns:
            List of ChunkIndexRecords sorted by message_start_time
        """
        # If we have already searched through the data section, there is nothing else
        # we can do to find more schema records, so we return what we have.
        if self._has_built_summary or self._has_searched_chunks:
            return self._cached_chunk_indexes

        # If reconstruction not allowed, return empty
        if self._enable_reconstruction == 'never':
            logging.warning('No chunk index records found in summary and reconstruction is disabled.')
            self._cached_chunk_indexes = []
            return self._cached_chunk_indexes

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Chunk indexes: file not initialized!')
            return self._cached_chunk_indexes

        # Search for chunk index records in the summary section
        logging.warning('No chunk index records found in summary. Searching through file!')
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)
        while McapRecordParser.peek_record(self._file) != McapRecordType.DATA_END:
            next_record = McapRecordParser.peek_record(self._file)
            if next_record == McapRecordType.CHUNK:
                chunk_start_offset = self._file.tell()
                chunk = McapRecordParser.parse_chunk(self._file)
                chunk_length = self._file.tell() - chunk_start_offset

                # Read the message index records
                message_index_length = 0
                reconstruct_message_index = True
                message_index_offsets: dict[int, int] = {}
                if self._enable_reconstruction != 'always':  # i.e. if 'missing'
                    start_post_chunk_message_index_offset = self._file.tell()
                    while McapRecordParser.peek_record(self._file) == McapRecordType.MESSAGE_INDEX:
                        post_chunk_message_index_offset = self._file.tell()
                        post_chunk_message_index = McapRecordParser.parse_message_index(self._file)
                        message_index_offsets[post_chunk_message_index.channel_id] = post_chunk_message_index_offset
                    message_index_length = self._file.tell() - start_post_chunk_message_index_offset
                    # If message index records exist, assume they are complete
                    reconstruct_message_index = not message_index_offsets
                    if reconstruct_message_index:
                        logging.warning("No message indexes found for chunk!")

                # Continue parsing the chunk
                compression = chunk.compression
                chunk_message_start_time: int | None = None
                chunk_message_end_time: int | None = None
                chunk_message_indexes: dict[ChannelId, list[tuple[LogTime, Offset]]] = defaultdict(list)

                compressed_size = len(chunk.records)
                reader = BytesReader(decompress_chunk(chunk, check_crc=self._check_crc))
                uncompressed_size = reader.size()

                while chunk_record_type := McapRecordParser.peek_record(reader):
                    if chunk_record_type == McapRecordType.MESSAGE:
                        message_offset = reader.tell()
                        chunk_message = McapRecordParser.parse_message(reader)

                        channel_id = chunk_message.channel_id
                        log_time = chunk_message.log_time
                        if reconstruct_message_index:
                            chunk_message_indexes[channel_id].append((log_time, message_offset))

                        if chunk_message_start_time is None or log_time < chunk_message_start_time:
                            chunk_message_start_time = log_time
                        if chunk_message_end_time is None or log_time > chunk_message_end_time:
                            chunk_message_end_time = log_time
                    else:
                        McapRecordParser.skip_record(reader)

                self._message_indexes[chunk_start_offset] = {
                    channel_id: MessageIndexRecord(
                        channel_id=channel_id,
                        records=sorted(records, key=lambda x: (x[0], x[1])),
                    )
                    for channel_id, records in chunk_message_indexes.items()
                }

                self._cached_chunk_indexes.append(
                    ChunkIndexRecord(
                        message_start_time=chunk_message_start_time or 0,
                        message_end_time=chunk_message_end_time or 0,
                        chunk_start_offset=chunk_start_offset,
                        chunk_length=chunk_length,
                        message_index_offsets=message_index_offsets,
                        message_index_length=message_index_length,
                        compression=compression,
                        compressed_size=compressed_size,
                        uncompressed_size=uncompressed_size,
                    )
                )
            else:
                McapRecordParser.skip_record(self._file)

        # Sort chunk indexes by message start time
        # TODO: Is this the best place to do this?
        self._cached_chunk_indexes.sort(key=lambda ci: ci.message_start_time)

        self._has_searched_chunks = True
        return self._cached_chunk_indexes

    def get_message_indexes(
        self,
        chunk_index: ChunkIndexRecord,
    ) -> dict[ChannelId, MessageIndexRecord]:
        """Get reconstructed message indexes for a chunk if available."""
        chunk_offset = chunk_index.chunk_start_offset
        if chunk_offset in self._message_indexes:
            return self._message_indexes[chunk_offset]

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Message indexes: file not initialized!')
            return {}

        current_pos = self._file.tell()
        try:
            _ = self._file.seek_from_start(chunk_offset)
            chunk = McapRecordParser.parse_chunk(self._file)
        finally:
            _ = self._file.seek_from_start(current_pos)

        reader = BytesReader(decompress_chunk(chunk, check_crc=self._check_crc))
        chunk_message_indexes: dict[ChannelId, list[tuple[int, int]]] = defaultdict(list)
        while chunk_record_type := McapRecordParser.peek_record(reader):
            if chunk_record_type == McapRecordType.MESSAGE:
                message_offset = reader.tell()
                message = McapRecordParser.parse_message(reader)
                chunk_message_indexes[message.channel_id].append((message.log_time, message_offset))
            else:
                McapRecordParser.skip_record(reader)

        reconstructed = {
            channel_id: MessageIndexRecord(
                channel_id=channel_id,
                records=sorted(records, key=lambda x: (x[0], x[1])),
            )
            for channel_id, records in chunk_message_indexes.items()
        }
        self._message_indexes[chunk_offset] = reconstructed

        return reconstructed

    def get_statistics(self) -> StatisticsRecord | None:
        """Get statistics about the MCAP file.

        Uses lazy loading: checks cache first, then loads from summary if available,
        or calculates from chunk indexes, otherwise returns pre-built data.

        Returns:
            StatisticsRecord containing file statistics
        """
        # If we have the statistics record return it, going through the data section
        # is dangerous to do because we might overwrite stuff we have added.
        if self._cached_statistics is not None:
            return self._cached_statistics

        # If reconstruction is not allowed, then nothing we can do
        if self._enable_reconstruction == 'never':
            logging.warning('No statistics record found in summary and reconstruction is disabled.')
            return None

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Statistics: file not initialized!')
            return None

        # Track message statistics
        chunk_count = 0
        message_count = 0
        attachment_count = 0
        metadata_count = 0
        message_start_time: int | None = None
        message_end_time: int | None = None
        channel_message_counts: dict[ChannelId, int] = defaultdict(int)
        schema_ids: set[SchemaId] = set()
        channel_ids: set[ChannelId] = set()

        # Seek to start of data section (after magic bytes)
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)

        # Iterate through records until we reach the footer
        logging.warning('No statistics record found in summary. Recreating from data section!')
        while McapRecordParser.peek_record(self._file) != McapRecordType.FOOTER:
            record_type = McapRecordParser.peek_record(self._file)
            if record_type == McapRecordType.SCHEMA:
                if schema := McapRecordParser.parse_schema(self._file):
                    schema_ids.add(schema.id)
            elif record_type == McapRecordType.CHANNEL:
                channel = McapRecordParser.parse_channel(self._file)
                channel_ids.add(channel.id)
            elif record_type == McapRecordType.CHUNK:
                chunk_count += 1
                chunk_start_offset = self._file.tell()
                chunk = McapRecordParser.parse_chunk(self._file)
                reader = BytesReader(decompress_chunk(chunk, check_crc=self._check_crc))
                chunk_message_indexes: dict[ChannelId, list[tuple[int, int]]] = defaultdict(list)
                while chunk_record_type := McapRecordParser.peek_record(reader):
                    if chunk_record_type == McapRecordType.SCHEMA:
                        if schema := McapRecordParser.parse_schema(reader):
                            schema_ids.add(schema.id)
                    elif chunk_record_type == McapRecordType.CHANNEL:
                        channel = McapRecordParser.parse_channel(reader)
                        channel_ids.add(channel.id)
                    elif chunk_record_type == McapRecordType.MESSAGE:
                        message_offset = reader.tell()
                        chunk_message = McapRecordParser.parse_message(reader)
                        channel_id = chunk_message.channel_id
                        log_time = chunk_message.log_time
                        channel_ids.add(channel_id)
                        message_count += 1
                        channel_message_counts[channel_id] += 1
                        chunk_message_indexes[channel_id].append((log_time, message_offset))
                        if message_start_time is None or log_time < message_start_time:
                            message_start_time = log_time
                        if message_end_time is None or log_time > message_end_time:
                            message_end_time = log_time
                    else:
                        McapRecordParser.skip_record(reader)

                if chunk_message_indexes and chunk_start_offset not in self._message_indexes:
                    self._message_indexes[chunk_start_offset] = {
                        channel_id: MessageIndexRecord(
                            channel_id=channel_id,
                            records=sorted(records, key=lambda x: (x[0], x[1])),
                        )
                        for channel_id, records in chunk_message_indexes.items()
                    }

                # Advance past any existing message index records following the chunk
                while McapRecordParser.peek_record(self._file) == McapRecordType.MESSAGE_INDEX:
                    McapRecordParser.skip_record(self._file)
            elif record_type == McapRecordType.MESSAGE:
                logging.warning('Found Message record outside of a chunk! Ignoring...')
                McapRecordParser.skip_record(self._file)
            elif record_type == McapRecordType.ATTACHMENT:
                attachment_count += 1
                McapRecordParser.skip_record(self._file)
            elif record_type == McapRecordType.METADATA:
                metadata_count += 1
                McapRecordParser.skip_record(self._file)
            else:
                McapRecordParser.skip_record(self._file)

        # Create statistics record from collected data
        self._cached_statistics = StatisticsRecord(
            message_count=message_count,
            schema_count=len(schema_ids),
            channel_count=len(channel_ids),
            attachment_count=attachment_count,
            metadata_count=metadata_count,
            chunk_count=chunk_count,
            message_start_time=message_start_time or 0,
            message_end_time=message_end_time or 0,
            channel_message_counts=dict(channel_message_counts),
        )
        return self._cached_statistics

    def get_attachment_indexes(self) -> dict[str, list[AttachmentIndexRecord]]:
        """Get all attachment indexes from the MCAP file.

        Returns:
            List of AttachmentIndexRecord objects.
        """
        # If we have already searched through the data section, there is nothing else
        # we can do to find more schema records, so we return what we have.
        if self._has_built_summary or self._has_searched_schemas:
            return self._cached_attachment_indexes

        # If reconstruction not allowed, return empty
        if self._enable_reconstruction == 'never':
            logging.warning('No attachment index records found in summary and reconstruction is disabled.')
            self._cached_attachment_indexes = {}
            return self._cached_attachment_indexes

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Attachment indexes: file not initialized!')
            return self._cached_attachment_indexes

        # Search for channel records in the summary section
        logging.warning('No attachment index records found in summary. Searching through file!')
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)
        while McapRecordParser.peek_record(self._file) != McapRecordType.DATA_END:
            next_record = McapRecordParser.peek_record(self._file)
            if next_record == McapRecordType.ATTACHMENT:
                start_offset = self._file.tell()
                attachment_record = McapRecordParser.parse_attachment(self._file)
                attachment_index = AttachmentIndexRecord(
                    offset=start_offset,
                    length=self._file.tell() - start_offset,
                    log_time=attachment_record.log_time,
                    create_time=attachment_record.create_time,
                    data_size=len(attachment_record.data),
                    name=attachment_record.name,
                    media_type=attachment_record.media_type,
                )
                self._cached_attachment_indexes[attachment_record.name].append(attachment_index)
            else:
                McapRecordParser.skip_record(self._file)
        self._has_searched_attachments = True  # Prevent searching through data section again
        return self._cached_attachment_indexes

    def get_metadata_indexes(self) -> dict[str, list[MetadataIndexRecord]]:
        """Get all metadata indexes from the MCAP file.

        Returns:
            List of MetadataIndexRecord objects.
        """
        # If we have already searched through the data section, there is nothing else
        # we can do to find more schema records, so we return what we have.
        if self._has_built_summary or self._has_searched_metadata:
            return self._cached_metadata_indexes

        # If reconstruction not allowed, return empty
        if self._enable_reconstruction == 'never':
            logging.warning('No metadata index records found in summary and reconstruction is disabled.')
            self._cached_metadata_indexes = {}
            return self._cached_metadata_indexes

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Metadata indexes: file not initialized!')
            return self._cached_metadata_indexes

        # Search for channel records in the summary section
        logging.warning('No attachment index records found in summary. Searching through file!')
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)
        while McapRecordParser.peek_record(self._file) != McapRecordType.DATA_END:
            next_record = McapRecordParser.peek_record(self._file)
            if next_record == McapRecordType.METADATA:
                start_offset = self._file.tell()
                metadata_record = McapRecordParser.parse_metadata(self._file)
                metadata_index = MetadataIndexRecord(
                    offset=start_offset,
                    length=self._file.tell() - start_offset,
                    name=metadata_record.name,
                )
                self._cached_metadata_indexes[metadata_record.name].append(metadata_index)
            else:
                McapRecordParser.skip_record(self._file)
        self._has_searched_metadata = True  # Prevent searching through data section again
        return self._cached_metadata_indexes

    def write_summary(self, writer: CrcWriter):
        # Write DataEnd record
        data_end = DataEndRecord(data_section_crc=writer.get_crc())
        McapRecordWriter.write_data_end(writer, data_end)

        # Write summary section using shared helper
        summary_start, summary_offset_start = _write_summary_section(
            writer,
            schema_records=list(self._cached_schemas.values()),
            channel_records=list(self._cached_channels.values()),
            statistics_record=self._cached_statistics,
            attachment_indexes=self._cached_attachment_indexes,
            chunk_indexes=self._cached_chunk_indexes,
            metadata_indexes=self._cached_metadata_indexes
        )

        # Write footer record manually for CRC calculation
        # TODO: Find a more elegant solution
        writer.write(McapRecordWriter._encode_record_type(McapRecordType.FOOTER))
        writer.write(McapRecordWriter._encode_uint64(FOOTER_PAYLOAD_SIZE))
        writer.write(McapRecordWriter._encode_uint64(summary_start))
        writer.write(McapRecordWriter._encode_uint64(summary_offset_start))
        writer.write(McapRecordWriter._encode_uint32(writer.get_crc()))

        # Write magic bytes again
        McapRecordWriter.write_magic_bytes(writer)


class McapNonChunkedSummary(McapSummary):
    """Summary information for a non-chunked MCAP file.

    This class handles loading summary information from a non-chunked MCAP file,
    either from existing summary sections or by reconstructing it from the data section.

    Args:
        file: The file reader to read from.
        enable_crc_check: Whether to validate the CRC checksums
        enable_reconstruction: Controls reconstruction behavior:
            - 'never': Raise error if summary sections are missing
            - 'missing': Load from summary if present, otherwise reconstruct
            - 'always': Always reconstruct even if summary exists
    """

    def __init__(
        self,
        file: BaseReader | None = None,
        *,
        enable_crc_check: bool = False,
        enable_reconstruction: Literal['never', 'missing', 'always'] = 'missing',
        load_summary_eagerly: bool = True,
    ) -> None:
        logging.debug('Creating McapNonChunkedSummary')
        self._file = file
        self._enable_reconstruction = enable_reconstruction
        self._check_crc = enable_crc_check
        self._load_summary_eagerly = load_summary_eagerly

        # Initialize cache variables for lazy loading
        self._summary_offset: dict[RecordId, Offset] = {}
        self._cached_schemas: dict[SchemaId, SchemaRecord] = {}
        self._cached_channels: dict[ChannelId, ChannelRecord] = {}
        self._cached_metadata_indexes: dict[str, list[MetadataIndexRecord]] = defaultdict(list)
        self._cached_attachment_indexes: dict[str, list[AttachmentIndexRecord]] = defaultdict(list)
        # If self._file is None, then we start a new summary and so we can track stats from the beginning
        self._cached_statistics: StatisticsRecord | None = StatisticsRecord() if self._file is None else None

        # If we have no file, then we have the summary built (i.e. nothing)
        self._has_built_summary = self._file is None
        self._has_loaded_summary = False

        self._has_searched_schemas = False
        self._has_searched_channels = False
        self._has_searched_attachments = False
        self._has_searched_metadata = False

        # If no file, then we are creating a new summary
        if self._file is None:
            logging.debug('Summary has no file to search')
            return

        logging.debug(f'Tell: {self._file.tell()}')

        _ = self._file.seek_from_end(FOOTER_SIZE + MAGIC_BYTES_SIZE)
        self._footer: FooterRecord = McapRecordParser.parse_footer(self._file)

        self._has_summary: bool = self._footer.summary_start != 0
        self._has_summary_offset: bool = self._footer.summary_offset_start != 0

        # Validate CRCs if requested
        if enable_crc_check:
            assert_data_crc(self._file, self._footer)
            assert_summary_crc(self._file, self._footer)

        # TODO: Figure out how to use offset
        if enable_reconstruction == 'never':
            if not self._has_summary:
                error_msg = 'No summary section detected in MCAP'
                raise McapNoSummarySectionError(error_msg)

        if enable_reconstruction == 'always' or not self._has_summary:
            logging.debug('Building summary from data section')
            self._build_summary()
            self._has_built_summary = True
        else:
            # TODO: What do we do if the summary is incomplete?
            logging.debug('Loading summary from summary section')
            self._load_summary()
            if self._cached_statistics is None and self._load_summary_eagerly:
                logging.warning('Statistics record not found in summary, generating')
                self._cached_statistics = self.get_statistics()

        # After loading summary, if we have records, then we assume they are complete
        # and we disable searching through the file for more records of the same type
        self._has_searched_schemas = len(self._cached_schemas) > 0
        self._has_searched_channels = len(self._cached_channels) > 0
        self._has_searched_attachments = len(self._cached_attachment_indexes) > 0
        self._has_searched_metadata = len(self._cached_metadata_indexes) > 0

    def _load_summary(self) -> None:
        if self._file is None:
            logging.warning('Load summary: file not initialized!')
            return None

        # Seek to start of summary section
        _ = self._file.seek_from_start(self._footer.summary_start)

        # Iterate through records until we reach the footer
        while (record_type := McapRecordParser.peek_record(self._file)) != McapRecordType.FOOTER:
            # Exit when we reach the end of the summary section
            if record_type == McapRecordType.SUMMARY_OFFSET:
                break

            if record_type == McapRecordType.SCHEMA:
                if schema := McapRecordParser.parse_schema(self._file):
                    self._cached_schemas[schema.id] = schema
            elif record_type == McapRecordType.CHANNEL:
                channel = McapRecordParser.parse_channel(self._file)
                self._cached_channels[channel.id] = channel
            elif record_type == McapRecordType.ATTACHMENT_INDEX:
                attachment_index = McapRecordParser.parse_attachment_index(self._file)
                self._cached_attachment_indexes[attachment_index.name].append(attachment_index)
            elif record_type == McapRecordType.METADATA_INDEX:
                metadata_index = McapRecordParser.parse_metadata_index(self._file)
                self._cached_metadata_indexes[metadata_index.name].append(metadata_index)
            elif record_type == McapRecordType.STATISTICS:
                self._cached_statistics = McapRecordParser.parse_statistics(self._file)
            elif record_type == McapRecordType.CHUNK_INDEX:
                logging.warning('Found CHUNK_INDEX in a non-chunked mcap!')
                McapRecordParser.skip_record(self._file)
            else:
                logging.error(f'Unexpected record in summary: {record_type}')
                McapRecordParser.skip_record(self._file)

    def _build_summary(self) -> None:
        """Build summary information by scanning the data section."""
        if self._file is None:
            logging.warning('Build summary: file not initialized!')
            return None

        # Track message statistics
        message_count = 0
        attachment_count = 0
        metadata_count = 0
        message_start_time: int | None = None
        message_end_time: int | None = None
        channel_message_counts: dict[ChannelId, int] = defaultdict(int)

        found_schemas: dict[SchemaId, SchemaRecord] = {}
        found_channels: dict[ChannelId, ChannelRecord]= {}

        # Seek to start of data section (after magic bytes)
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)

        # Iterate through records until we reach the footer
        while (record_type := McapRecordParser.peek_record(self._file)) != McapRecordType.FOOTER:
            if record_type == McapRecordType.SCHEMA:
                if schema := McapRecordParser.parse_schema(self._file):
                    found_schemas[schema.id] = schema
            elif record_type == McapRecordType.CHANNEL:
                channel = McapRecordParser.parse_channel(self._file)
                found_channels[channel.id] = channel
            elif record_type == McapRecordType.MESSAGE:
                # Parse message to extract statistics
                message = McapRecordParser.parse_message(self._file)
                channel_id = message.channel_id
                log_time = message.log_time

                # Update statistics
                message_count += 1
                channel_message_counts[channel_id] += 1
                if message_start_time is None or log_time < message_start_time:
                    message_start_time = log_time
                if message_end_time is None or log_time > message_end_time:
                    message_end_time = log_time
            elif record_type == McapRecordType.ATTACHMENT:
                attachment_count += 1
                start_offset = self._file.tell()
                attachment_record = McapRecordParser.parse_attachment(self._file)
                attachment_index = AttachmentIndexRecord(
                    offset=start_offset,
                    length=self._file.tell() - start_offset,
                    log_time=attachment_record.log_time,
                    create_time=attachment_record.create_time,
                    data_size=len(attachment_record.data),
                    name=attachment_record.name,
                    media_type=attachment_record.media_type,
                )
                self._cached_attachment_indexes[attachment_record.name].append(attachment_index)
            elif record_type == McapRecordType.METADATA:
                metadata_count += 1
                start_offset = self._file.tell()
                metadata_record = McapRecordParser.parse_metadata(self._file)
                metadata_index = MetadataIndexRecord(
                    offset=start_offset,
                    length=self._file.tell() - start_offset,
                    name=metadata_record.name,
                )
                self._cached_metadata_indexes[metadata_record.name].append(metadata_index)
            else:
                McapRecordParser.skip_record(self._file)

        # Create statistics record from collected data
        self._cached_statistics = StatisticsRecord(
            message_count=message_count,
            schema_count=len(found_schemas),
            channel_count=len(found_channels),
            attachment_count=attachment_count,
            metadata_count=metadata_count,
            chunk_count=0,  # Non-chunked files have no chunks
            message_start_time=message_start_time or 0,
            message_end_time=message_end_time or 0,
            channel_message_counts=channel_message_counts,
        )
        self._cached_schemas = found_schemas
        self._cached_channels = found_channels

    def next_schema_id(self) -> int:
        # Schema ID must be non-zero
        return max(self.get_schemas().keys(), default=0) + 1

    def next_channel_id(self) -> int:
        # No harm in starting channel IDs from 1 too
        return max(self.get_channels().keys(), default=0) + 1

    def next_sequence_id(self, channel_id: ChannelId) -> int:
        if self._cached_statistics is not None:
            # We should always start at 1 because 0 indicates "not relevant"
            return self._cached_statistics.channel_message_counts.get(channel_id, 0) + 1
        return 0

    def get_channel_id(self, topic: str) -> ChannelId | None:
        for id, record in self._cached_channels.items():
            if record.topic == topic:
                return id
        return None

    def get_schema_id(self, message: type[Message]) -> SchemaId | None:
        for id, record in self._cached_schemas.items():
            if record.name == message.__msg_name__:
                return id
        return None

    def add_schema(self, schema: SchemaRecord):
        if schema.id in self._cached_schemas:
            logging.warning(f'Schema (id {schema.id}) already written to file')
        else:
            self._cached_schemas[schema.id] = schema
            if self._cached_statistics is not None:
                self._cached_statistics.schema_count += 1

    def add_channel(self, channel: ChannelRecord):
        if channel.id in self._cached_channels:
            logging.warning(f'Channel (id {channel.id}) already written to file')
        else:
            self._cached_channels[channel.id] = channel
            if self._cached_statistics is not None:
                self._cached_statistics.channel_count += 1
                self._cached_statistics.channel_message_counts[channel.id] = 0

    def add_message(self, message: MessageRecord):
        if self._cached_statistics is not None:
            self._cached_statistics.message_count += 1
            self._cached_statistics.channel_message_counts[message.channel_id] += 1
            # TODO: Find something more elegant
            self._cached_statistics.message_start_time = min(
                message.log_time if self._cached_statistics.message_start_time <= 0 else self._cached_statistics.message_start_time,
                message.log_time,
            )
            self._cached_statistics.message_end_time = max(
                message.log_time if self._cached_statistics.message_end_time <= 0 else self._cached_statistics.message_end_time,
                message.log_time,
            )

    def add_attachment_index(self, attachment_index: AttachmentIndexRecord):
        self._cached_attachment_indexes[attachment_index.name].append(attachment_index)
        if self._cached_statistics is not None:
            self._cached_statistics.attachment_count += 1

    def add_metadata_index(self, metadata_index: MetadataIndexRecord):
        self._cached_metadata_indexes[metadata_index.name].append(metadata_index)
        if self._cached_statistics is not None:
            self._cached_statistics.metadata_count += 1

    def get_schemas(self) -> dict[SchemaId, SchemaRecord]:
        """Get all schemas defined in the MCAP file."""
        # If we have already searched through the data section, there is nothing else
        # we can do to find more schema records, so we return what we have.
        if self._has_built_summary or self._has_searched_schemas:
            return self._cached_schemas

        # Finally, we have yet to search through the data section to find schemas
        # If reconstruction not allowed, we cannot do anything so we return what we have
        if self._enable_reconstruction == 'never':
            logging.warning('No schema records found in summary and reconstruction is disabled.')
            return self._cached_schemas

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Schemas: file not initialized!')
            return self._cached_schemas

        # We are allowed to search through the data section for any schemas that we can find
        logging.warning('No schema records found in summary. Searching through file!')
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)
        while (next_record := McapRecordParser.peek_record(self._file)) != McapRecordType.DATA_END:
            if next_record != McapRecordType.SCHEMA:
                McapRecordParser.skip_record(self._file)
                continue
            if record := McapRecordParser.parse_schema(self._file):
                self._cached_schemas[record.id] = record
        self._has_searched_schemas = True  # Prevent searching through data section again
        return self._cached_schemas

    def get_channels(self) -> dict[ChannelId, ChannelRecord]:
        """Get all channels defined in the MCAP file."""
        # If we have already searched through the data section, there is nothing else
        # we can do to find more channel records, so we return what we have.
        if self._has_built_summary or self._has_searched_channels:
            return self._cached_channels

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Channels: file not initialized!')
            return self._cached_channels

        # Finally, we have yet to search through the data section to find channels
        # If reconstruction not allowed, we cannot do anything so we return what we have
        if self._enable_reconstruction == 'never':
            logging.warning('No channel records found in summary and reconstruction is disabled.')
            return self._cached_channels

        # We are allowed to search through the data section for any schemas that we can find
        logging.warning('No channel records found in summary. Searching through file!')
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)
        while (next_record := McapRecordParser.peek_record(self._file)) != McapRecordType.DATA_END:
            if next_record != McapRecordType.CHANNEL:
                McapRecordParser.skip_record(self._file)
                continue
            if record := McapRecordParser.parse_channel(self._file):
                self._cached_channels[record.id] = record
        self._has_searched_channels = True  # Prevent searching through data section again
        return self._cached_channels

    def get_chunk_indexes(self) -> list[ChunkIndexRecord]:
        return []

    def get_message_indexes(
        self,
        chunk_index: ChunkIndexRecord,
    ) -> dict[ChannelId, MessageIndexRecord]:
        return {}  # TODO: Move from McapNonChunkedReader

    def get_statistics(self) -> StatisticsRecord | None:
        """Get statistics about the MCAP file."""
        # If we have the statistics record return it, going through the data section
        # is dangerous to do because we might overwrite stuff we have added.
        if self._cached_statistics is not None:
            return self._cached_statistics

        # If reconstruction is not allowed, then nothing we can do
        if self._enable_reconstruction == 'never':
            logging.warning('No statistics record found in summary and reconstruction is disabled.')
            return None

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Statistics: file not initialized!')
            return None

        # We use None to indicate that we have not yet searched through the data section
        # because once we have, the cache will no longer be None.

        # Track message statistics
        schema_count = 0
        channel_count = 0
        message_count = 0
        attachment_count = 0
        metadata_count = 0
        message_start_time: int | None = None
        message_end_time: int | None = None
        channel_message_counts: dict[ChannelId, int] = defaultdict(int)

        # Seek to start of data section (after magic bytes)
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)

        # Iterate through records until we reach the footer
        logging.warning('No statistics record found in summary. Recreating from data section!')
        while (record_type := McapRecordParser.peek_record(self._file)) != McapRecordType.FOOTER:
            if record_type == McapRecordType.SCHEMA:
                schema_count += 1
                McapRecordParser.skip_record(self._file)
            elif record_type == McapRecordType.CHANNEL:
                channel_count += 1
                McapRecordParser.skip_record(self._file)
            elif record_type == McapRecordType.MESSAGE:
                # Parse message to extract statistics
                message = McapRecordParser.parse_message(self._file)
                channel_id = message.channel_id
                log_time = message.log_time

                # Update statistics
                message_count += 1
                channel_message_counts[channel_id] += 1
                if message_start_time is None or log_time < message_start_time:
                    message_start_time = log_time
                if message_end_time is None or log_time > message_end_time:
                    message_end_time = log_time
            elif record_type == McapRecordType.ATTACHMENT:
                attachment_count += 1
                McapRecordParser.skip_record(self._file)
            elif record_type == McapRecordType.METADATA:
                metadata_count += 1
                McapRecordParser.skip_record(self._file)
            else:
                McapRecordParser.skip_record(self._file)

        # Create statistics record from collected data
        self._cached_statistics = StatisticsRecord(
            message_count=message_count,
            schema_count=schema_count,
            channel_count=channel_count,
            attachment_count=attachment_count,
            metadata_count=metadata_count,
            chunk_count=0,  # Non-chunked files have no chunks
            message_start_time=message_start_time or 0,
            message_end_time=message_end_time or 0,
            channel_message_counts=channel_message_counts,
        )
        return self._cached_statistics

    def get_attachment_indexes(self) -> dict[str, list[AttachmentIndexRecord]]:
        """Get all attachment indexes from the MCAP file.

        Returns:
            List of AttachmentIndexRecord objects.
        """
        # If we have already searched through the data section, there is nothing else
        # we can do to find more schema records, so we return what we have.
        if self._has_built_summary or self._has_searched_attachments:
            return self._cached_attachment_indexes

        # If reconstruction not allowed, there is nothing we can do
        if self._enable_reconstruction == 'never':
            logging.warning('No attachment index records found in summary and reconstruction is disabled.')
            return self._cached_attachment_indexes

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Attachment indexes: file not initialized!')
            return self._cached_attachment_indexes

        # Search for channel records in the summary section
        logging.warning('No attachment index records found in summary. Searching through file!')
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)
        while McapRecordParser.peek_record(self._file) != McapRecordType.DATA_END:
            next_record = McapRecordParser.peek_record(self._file)
            if next_record == McapRecordType.ATTACHMENT:
                start_offset = self._file.tell()
                attachment_record = McapRecordParser.parse_attachment(self._file)
                attachment_index = AttachmentIndexRecord(
                    offset=start_offset,
                    length=self._file.tell() - start_offset,
                    log_time=attachment_record.log_time,
                    create_time=attachment_record.create_time,
                    data_size=len(attachment_record.data),
                    name=attachment_record.name,
                    media_type=attachment_record.media_type,
                )
                self._cached_attachment_indexes[attachment_record.name].append(attachment_index)
            else:
                McapRecordParser.skip_record(self._file)
        self._has_searched_attachments = True  # Prevent searching through data section again
        return self._cached_attachment_indexes

    def get_metadata_indexes(self) -> dict[str, list[MetadataIndexRecord]]:
        """Get all metadata indexes from the MCAP file.

        Returns:
            List of MetadataIndexRecord objects.
        """
        # If we have already searched through the data section, there is nothing else
        # we can do to find more schema records, so we return what we have.
        if self._has_built_summary or self._has_searched_metadata:
            return self._cached_metadata_indexes

        # If reconstruction not allowed, there is nothing we can do
        if self._enable_reconstruction == 'never':
            logging.warning('No metadata index records found in summary and reconstruction is disabled.')
            return self._cached_metadata_indexes

        # If no file is given, we cannot search through the data
        if self._file is None:
            logging.warning('Metadata indexes: file not initialized!')
            return self._cached_metadata_indexes

        # Search for channel records in the summary section
        logging.warning('No attachment index records found in summary. Searching through file!')
        _ = self._file.seek_from_start(MAGIC_BYTES_SIZE)
        while McapRecordParser.peek_record(self._file) != McapRecordType.DATA_END:
            next_record = McapRecordParser.peek_record(self._file)
            if next_record == McapRecordType.METADATA:
                start_offset = self._file.tell()
                metadata_record = McapRecordParser.parse_metadata(self._file)
                metadata_index = MetadataIndexRecord(
                    offset=start_offset,
                    length=self._file.tell() - start_offset,
                    name=metadata_record.name,
                )
                self._cached_metadata_indexes[metadata_record.name].append(metadata_index)
            else:
                McapRecordParser.skip_record(self._file)
        self._has_searched_metadata = True  # Prevent searching through data section again
        return self._cached_metadata_indexes

    def write_summary(self, writer: CrcWriter):
        # Write DataEnd record
        data_end = DataEndRecord(data_section_crc=writer.get_crc())
        McapRecordWriter.write_data_end(writer, data_end)

        # Write summary section using shared helper
        summary_start, summary_offset_start = _write_summary_section(
            writer,
            schema_records=list(self._cached_schemas.values()),
            channel_records=list(self._cached_channels.values()),
            statistics_record=self._cached_statistics,
            attachment_indexes=self._cached_attachment_indexes,
            metadata_indexes=self._cached_metadata_indexes
        )

        # Write footer record manually for CRC calculation
        writer.write(McapRecordWriter._encode_record_type(McapRecordType.FOOTER))
        writer.write(McapRecordWriter._encode_uint64(FOOTER_PAYLOAD_SIZE))
        writer.write(McapRecordWriter._encode_uint64(summary_start))
        writer.write(McapRecordWriter._encode_uint64(summary_offset_start))
        writer.write(McapRecordWriter._encode_uint32(writer.get_crc()))

        # Write magic bytes again
        McapRecordWriter.write_magic_bytes(writer)


class McapSummaryFactory:
    @staticmethod
    def create_summary(
        file: BaseReader | None = None,
        chunk_size: int | None = None,
        enable_reconstruction: Literal['never', 'missing', 'always'] = 'missing',
        load_summary_eagerly: bool = True,
    ) -> McapSummary:
        if file:  # Append mode
            try:
                return McapChunkedSummary(
                    file,
                    enable_reconstruction=enable_reconstruction,
                    load_summary_eagerly=load_summary_eagerly,
                )
            except McapNoChunkIndexError:
                return McapNonChunkedSummary(
                    file,
                    enable_reconstruction=enable_reconstruction,
                    load_summary_eagerly=load_summary_eagerly,
                )
        elif chunk_size is None:
            return McapNonChunkedSummary(
                enable_reconstruction=enable_reconstruction,
                load_summary_eagerly=load_summary_eagerly,
            )
        return McapChunkedSummary(
            enable_reconstruction=enable_reconstruction,
            load_summary_eagerly=load_summary_eagerly,
        )


def _write_summary_section(
    writer: CrcWriter,
    schema_records: list[SchemaRecord] | None = None,
    channel_records: list[ChannelRecord] | None = None,
    statistics_record: StatisticsRecord | None = None,
    chunk_indexes: list[ChunkIndexRecord] | None = None,
    attachment_indexes: dict[str, list[AttachmentIndexRecord]] | None = None,
    metadata_indexes: dict[str, list[MetadataIndexRecord]] | None = None
) -> tuple[int, int]:
    """Write the summary section and return (summary_start, summary_offset_start).

    Args:
        chunk_indexes: Optional list of chunk index records (for chunked writers).
        attachment_indexes: Optional list of attachment index records.
        metadata_indexes: Optional list of metadata index records.

    Returns:
        Tuple of (summary_start, summary_offset_start) positions.
    """
    # Start summary section
    summary_start = writer.tell()
    writer.clear_crc()

    # Write schema records to summary
    schema_group_start = summary_start
    if schema_records:
        logging.debug(f'Writing {len(schema_records)} schema records')
        for record in schema_records:
            McapRecordWriter.write_schema(writer, record)
    schema_group_length = writer.tell() - schema_group_start

    # Write channel records to summary
    channel_group_start = writer.tell()
    if channel_records:
        logging.debug(f'Writing {len(channel_records)} channel records')
        for record in channel_records:
            McapRecordWriter.write_channel(writer, record)
    channel_group_length = writer.tell() - channel_group_start

    # Write attachment index records to summary
    attachment_index_group_start = writer.tell()
    if attachment_indexes:
        attachment_count = 0
        for record_list in attachment_indexes.values():
            for record in record_list:
                attachment_count += 1
                McapRecordWriter.write_attachment_index(writer, record)
        logging.debug(f'Writing {attachment_count} attachment index records')
    attachment_index_group_length = writer.tell() - attachment_index_group_start

    # Write metadata index records to summary
    metadata_index_group_start = writer.tell()
    if metadata_indexes:
        metadata_count = 0
        for record_list in metadata_indexes.values():
            for record in record_list:
                metadata_count += 1
                McapRecordWriter.write_metadata_index(writer, record)
        logging.debug(f'Writing {metadata_count} metadata index records')
    metadata_index_group_length = writer.tell() - metadata_index_group_start

    # Write chunk index records to summary (only for chunked writers)
    chunk_index_group_start = writer.tell()
    if chunk_indexes:
        logging.debug(f'Writing {len(chunk_indexes)} chunk index records')
        for record in chunk_indexes:
            McapRecordWriter.write_chunk_index(writer, record)
    chunk_index_group_length = writer.tell() - chunk_index_group_start

    # Write statistics record
    statistics_group_start = writer.tell()
    if statistics_record is not None:
        logging.debug(f'Writing statistics record')
        McapRecordWriter.write_statistics(writer, statistics_record)
    statistics_group_length = writer.tell() - statistics_group_start

    # Write summary offsets
    summary_offset_start = writer.tell()
    if schema_group_length > 0:
        McapRecordWriter.write_summary_offset(
            writer,
            SummaryOffsetRecord(
                group_opcode=RecordType.SCHEMA,
                group_start=schema_group_start,
                group_length=schema_group_length,
            ),
        )
    if channel_group_length > 0:
        McapRecordWriter.write_summary_offset(
            writer,
            SummaryOffsetRecord(
                group_opcode=RecordType.CHANNEL,
                group_start=channel_group_start,
                group_length=channel_group_length,
            ),
        )
    if attachment_index_group_length > 0:
        McapRecordWriter.write_summary_offset(
            writer,
            SummaryOffsetRecord(
                group_opcode=RecordType.ATTACHMENT_INDEX,
                group_start=attachment_index_group_start,
                group_length=attachment_index_group_length,
            ),
        )
    if metadata_index_group_length > 0:
        McapRecordWriter.write_summary_offset(
            writer,
            SummaryOffsetRecord(
                group_opcode=RecordType.METADATA_INDEX,
                group_start=metadata_index_group_start,
                group_length=metadata_index_group_length,
            ),
        )
    if chunk_index_group_length > 0:
        McapRecordWriter.write_summary_offset(
            writer,
            SummaryOffsetRecord(
                group_opcode=RecordType.CHUNK_INDEX,
                group_start=chunk_index_group_start,
                group_length=chunk_index_group_length,
            ),
        )
    if statistics_group_length > 0:
        McapRecordWriter.write_summary_offset(
            writer,
            SummaryOffsetRecord(
                group_opcode=RecordType.STATISTICS,
                group_start=statistics_group_start,
                group_length=statistics_group_length,
            ),
        )

    return summary_start, summary_offset_start



"""Tests for the MCAP reader."""
import logging
import random
from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Literal

import pytest
from mcap.reader import make_reader
from mcap_ros1.decoder import DecoderFactory as Ros1DecoderFactory
from mcap_ros2.decoder import DecoderFactory as Ros2DecoderFactory
from mcap_ros2.writer import Writer as Ros2McapWriter
from rosbags.rosbag2 import StoragePlugin, Writer
from rosbags.typesys import Stores, get_typestore
from rosbags.typesys.store import Typestore

import pybag.ros1.noetic.std_msgs as ros1_std_msgs
import pybag.ros2.humble.std_msgs as ros2_std_msgs
import pybag.types as t
from pybag.mcap_reader import McapFileReader, McapMultipleFileReader
from pybag.mcap_writer import McapFileWriter
from pybag.schema.ros2msg import Ros2MsgError


def _find_mcap_file(temp_dir: str) -> Path:
    return next(Path(temp_dir).rglob("*.mcap"))


@pytest.fixture(params=[Stores.ROS2_JAZZY, Stores.ROS2_HUMBLE])
def typestore(request):
    return get_typestore(request.param)

#################
#  Pybag tests  #
#################

@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
def test_messages_filter(typestore: Typestore, in_log_time_order: bool, enable_crc_check: bool):
    # Write a temporary mcap file
    Int32 = typestore.types["std_msgs/msg/Int32"]
    with TemporaryDirectory() as temp_dir:
        with Writer(
            Path(temp_dir) / "rosbags",
            version=9,
            storage_plugin=StoragePlugin.MCAP,
        ) as writer:
            conn = writer.add_connection("/rosbags", Int32.__msgtype__, typestore=typestore)
            writer.write(conn, 0, typestore.serialize_cdr(Int32(data=1), Int32.__msgtype__))
            writer.write(conn, 1, typestore.serialize_cdr(Int32(data=-1), Int32.__msgtype__))

        mcap_file = _find_mcap_file(temp_dir)
        with McapFileReader.from_file(mcap_file, enable_crc_check=enable_crc_check) as reader:
            all_messages = list(reader.messages("/rosbags", in_log_time_order=in_log_time_order))
            assert len(all_messages) == 2
            assert all_messages[0].data.data == 1
            assert all_messages[1].data.data == -1

            positive = list(reader.messages(
                "/rosbags",
                filter=lambda msg: msg.data.data > 0,
                in_log_time_order=in_log_time_order
            ))
            assert len(positive) == 1
            assert positive[0].data.data == 1

            negative = list(reader.messages(
                "/rosbags",
                filter=lambda msg: msg.data.data < 0,
                in_log_time_order=in_log_time_order
            ))
            assert len(negative) == 1
            assert negative[0].data.data == -1


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_multiple_existing_topics(chunk_size, in_log_time_order: bool, enable_crc_check: bool, profile):
    """Test reading messages from multiple existing topics."""
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "multiple_topics.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            # Write messages to /topic1 with timestamps 10, 20, 30
            writer.write_message("/topic1", 10, String(data="topic1_0"))
            writer.write_message("/topic1", 20, String(data="topic1_1"))
            writer.write_message("/topic1", 30, String(data="topic1_2"))

            # Write messages to /topic2 with timestamps 15, 25, 35
            writer.write_message("/topic2", 15, String(data="topic2_0"))
            writer.write_message("/topic2", 25, String(data="topic2_1"))
            writer.write_message("/topic2", 35, String(data="topic2_2"))

            # Write messages to /topic3 with timestamps 5, 45, 50
            writer.write_message("/topic3", 5, String(data="topic3_0"))
            writer.write_message("/topic3", 45, String(data="topic3_1"))
            writer.write_message("/topic3", 50, String(data="topic3_2"))

        # Read messages with pybag
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            # Check that we have multiple chunks if configured to do so
            chunk_indexes = reader._reader.get_chunk_indexes()
            logging.info(f'Number of chunks: {len(chunk_indexes)}')
            assert chunk_size is None or len(chunk_indexes) > 1, "Expected multiple chunks"

            messages = list(reader.messages(["/topic1", "/topic2"], in_log_time_order=in_log_time_order))
            logging.info(f'pybag: {[msg.log_time for msg in messages]}')
            logging.info(f'pybag: {[msg.data.data for msg in messages]}')

            # Assert that 6 messages are returned (3 from each topic)
            assert len(messages) == 6, f"Expected 6 messages, got {len(messages)}"

            if in_log_time_order:
                expected_log_times = [10, 15, 20, 25, 30, 35]
                expected_data = ["topic1_0", "topic2_0", "topic1_1", "topic2_1", "topic1_2", "topic2_2"]
            else:
                expected_log_times = [10, 20, 30, 15, 25, 35]
                expected_data = ["topic1_0", "topic1_1", "topic1_2", "topic2_0", "topic2_1", "topic2_2"]

            assert [msg.log_time for msg in messages] == expected_log_times, \
                f"Expected log times {expected_log_times}, got {[msg.log_time for msg in messages]}"
            assert [msg.data.data for msg in messages] == expected_data, \
                f"Expected data {expected_data}, got {[msg.data.data for msg in messages]}"

@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_multiple_topics_with_nonexistent(chunk_size, in_log_time_order: bool, enable_crc_check: bool, profile):
    """Test reading messages from multiple topics where some don't exist."""
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "partial_topics.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            # Write 10 messages to /existing to ensure multiple chunks are created
            for i in range(10):
                writer.write_message("/existing", i * 10, String(data=f"msg_{i}"))

        # Read messages with pybag
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            # Check that we have multiple chunks if configured to do so
            chunk_indexes = reader._reader.get_chunk_indexes()
            logging.info(f'Number of chunks: {len(chunk_indexes)}')
            assert chunk_size is None or len(chunk_indexes) > 1, "Expected multiple chunks"

            # Call with both an existing and non-existent topic
            messages = list(reader.messages(["/existing", "/nonexistent"], in_log_time_order=in_log_time_order))
            logging.info(f'pybag: {[msg.log_time for msg in messages]}')
            logging.info(f'pybag: {[msg.data.data for msg in messages]}')

            # Assert that only 10 messages are returned (from /existing only)
            assert len(messages) == 10, f"Expected 10 messages from /existing, got {len(messages)}"

            # Assert the log times and data are correct
            expected_log_times = list(range(0, 100, 10))
            expected_data = [f"msg_{i}" for i in range(10)]
            assert [msg.log_time for msg in messages] == expected_log_times, \
                f"Expected log times {expected_log_times}, got {[msg.log_time for msg in messages]}"
            assert [msg.data.data for msg in messages] == expected_data, \
                f"Expected data {expected_data}, got {[msg.data.data for msg in messages]}"


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_multiple_nonexistent_topics(chunk_size, in_log_time_order: bool, enable_crc_check: bool, profile):
    """Test reading messages from multiple topics that don't exist."""
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "no_topics.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            # Write enough messages to /real_topic to create multiple chunks
            for i in range(10):
                writer.write_message("/real_topic", i * 10, String(data=f"msg_{i}"))

        # Read messages with pybag
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            # Check that we have multiple chunks if configured to do so
            chunk_indexes = reader._reader.get_chunk_indexes()
            logging.info(f'Number of chunks: {len(chunk_indexes)}')
            assert chunk_size is None or len(chunk_indexes) > 1, "Expected multiple chunks"

            # Call with non-existent topics
            messages = list(reader.messages(["/fake1", "/fake2"], in_log_time_order=in_log_time_order))
            logging.info(f'pybag: {[msg.log_time for msg in messages]}')
            logging.info(f'pybag: number of messages: {len(messages)}')

            # Assert that 0 messages are returned (empty list)
            assert len(messages) == 0, f"Expected 0 messages from non-existent topics, got {len(messages)}"


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_glob_pattern_matching(chunk_size, in_log_time_order: bool, enable_crc_check: bool, profile):
    """Test reading messages using glob patterns to match multiple topics.

    This test verifies that glob patterns like '/sensor/*' can be used to match
    multiple topics that share a common prefix. The test creates topics under
    two different namespaces (/sensor and /control) and validates:
    1. Glob patterns correctly match all topics under a namespace
    2. Messages are returned in the correct order (log_time or write order)
    3. Glob patterns with no matches return an empty list without errors
    """
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "glob_pattern.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            # /sensor/camera
            writer.write_message("/sensor/camera", 10, String(data="camera_0"))
            writer.write_message("/sensor/camera", 20, String(data="camera_1"))
            # /sensor/lidar
            writer.write_message("/sensor/lidar", 15, String(data="lidar_0"))
            writer.write_message("/sensor/lidar", 25, String(data="lidar_1"))
            # /sensor/imu
            writer.write_message("/sensor/imu", 12, String(data="imu_0"))
            writer.write_message("/sensor/imu", 22, String(data="imu_1"))
            # /control
            writer.write_message("/control/speed", 8, String(data="speed_0"))
            writer.write_message("/control/steering", 30, String(data="steering_0"))
            # Write second speed message AFTER steering (out of timestamp order)
            writer.write_message("/control/speed", 18, String(data="speed_1"))

        # Read messages with pybag
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            # Check that we have multiple chunks if configured to do so
            chunk_indexes = reader._reader.get_chunk_indexes()
            logging.info(f'Number of chunks: {len(chunk_indexes)}')
            assert chunk_size is None or len(chunk_indexes) > 1, "Expected multiple chunks"

            # Test case 1: Glob pattern matching /sensor/* topics
            messages = list(reader.messages("/sensor/*", in_log_time_order=in_log_time_order))
            logging.info(f'pybag /sensor/*: {[msg.log_time for msg in messages]}')
            logging.info(f'pybag /sensor/*: {[msg.data.data for msg in messages]}')
            assert len(messages) == 6, f"Expected 6 messages from /sensor/*, got {len(messages)}"

            if in_log_time_order:
                expected_log_times = [10, 12, 15, 20, 22, 25]
                expected_data = ["camera_0", "imu_0", "lidar_0", "camera_1", "imu_1", "lidar_1"]
            else:
                expected_log_times = [10, 20, 15, 25, 12, 22]
                expected_data = ["camera_0", "camera_1", "lidar_0", "lidar_1", "imu_0", "imu_1"]

            assert [msg.log_time for msg in messages] == expected_log_times, \
                f"Expected log times {expected_log_times}, got {[msg.log_time for msg in messages]}"
            assert [msg.data.data for msg in messages] == expected_data, \
                f"Expected data {expected_data}, got {[msg.data.data for msg in messages]}"

            # Test case 2: Glob pattern matching /control/* topics
            messages = list(reader.messages("/control/*", in_log_time_order=in_log_time_order))
            logging.info(f'pybag /control/*: {[msg.log_time for msg in messages]}')
            logging.info(f'pybag /control/*: {[msg.data.data for msg in messages]}')
            assert len(messages) == 3, f"Expected 3 messages from /control/*, got {len(messages)}"

            if in_log_time_order:
                expected_log_times = [8, 18, 30]
                expected_data = ["speed_0", "speed_1", "steering_0"]
            else:
                expected_log_times = [8, 30, 18]
                expected_data = ["speed_0", "steering_0", "speed_1"]

            assert [msg.log_time for msg in messages] == expected_log_times, \
                f"Expected log times {expected_log_times}, got {[msg.log_time for msg in messages]}"
            assert [msg.data.data for msg in messages] == expected_data, \
                f"Expected data {expected_data}, got {[msg.data.data for msg in messages]}"

            # Test case 3: Glob pattern with no matches
            messages = list(reader.messages("/nonexistent/*", in_log_time_order=in_log_time_order))
            logging.info(f'pybag /nonexistent/*: number of messages: {len(messages)}')
            assert len(messages) == 0, f"Expected 0 messages from /nonexistent/*, got {len(messages)}"


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_ordered_messages(chunk_size, in_log_time_order: bool, enable_crc_check: bool, profile):
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "unordered.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            writer.write_message("/unordered", 0, String(data=f"msg_{0}"))
            writer.write_message("/unordered", 1, String(data=f"msg_{1}"))
            writer.write_message("/unordered", 2, String(data=f"msg_{2}"))
            writer.write_message("/unordered", 3, String(data=f"msg_{3}"))
            writer.write_message("/unordered", 4, String(data=f"msg_{4}"))
            writer.write_message("/unordered", 5, String(data=f"msg_{5}"))
            writer.write_message("/unordered", 6, String(data=f"msg_{6}"))
            writer.write_message("/unordered", 7, String(data=f"msg_{7}"))

        # Read messages with official mcap library
        with open(path, 'rb') as f:
            decoder_factory = Ros2DecoderFactory() if profile == 'ros2' else Ros1DecoderFactory()
            reader = make_reader(f, decoder_factories=[decoder_factory])
            official_mcap_messages = list(reader.iter_decoded_messages(log_time_order=True))
            logging.info(f'mcap: {[msg[-2].log_time for msg in official_mcap_messages]}')

        # Read messages with pybag
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            # Check that we have multiple chunks if configured to do so
            chunk_indexes = reader._reader.get_chunk_indexes()
            logging.info(f'Number of chunks: {len(chunk_indexes)}')
            assert chunk_size is None or len(chunk_indexes) > 1, "Expected multiple chunks"

            messages = list(reader.messages("/unordered", in_log_time_order=in_log_time_order))
            logging.info(f'pybag: {[message.log_time for message in messages]}')
            logging.info(f'pybag: {[msg.data.data for msg in messages]}')

            expected_log_times = list(range(8))
            assert [msg.log_time for msg in messages] == expected_log_times
            assert [msg.data.data for msg in messages] == [f"msg_{t}" for t in expected_log_times]


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_reverse_ordered_messages(chunk_size, in_log_time_order: bool, enable_crc_check: bool, profile):
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "unordered.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            writer.write_message("/unordered", 7, String(data=f"msg_{7}"))
            writer.write_message("/unordered", 6, String(data=f"msg_{6}"))
            writer.write_message("/unordered", 5, String(data=f"msg_{5}"))
            writer.write_message("/unordered", 4, String(data=f"msg_{4}"))
            writer.write_message("/unordered", 3, String(data=f"msg_{3}"))
            writer.write_message("/unordered", 2, String(data=f"msg_{2}"))
            writer.write_message("/unordered", 1, String(data=f"msg_{1}"))
            writer.write_message("/unordered", 0, String(data=f"msg_{0}"))

        # Read messages with official mcap library
        with open(path, 'rb') as f:
            decoder_factory = Ros2DecoderFactory() if profile == 'ros2' else Ros1DecoderFactory()
            reader = make_reader(f, decoder_factories=[decoder_factory])
            official_mcap_messages = list(reader.iter_decoded_messages(log_time_order=True))
            logging.info(f'mcap: {[msg[-2].log_time for msg in official_mcap_messages]}')

        # Read messages with pybag
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            # Check that we have multiple chunks if configured to do so
            chunk_indexes = reader._reader.get_chunk_indexes()
            logging.info(f'Number of chunks: {len(chunk_indexes)}')
            assert chunk_size is None or len(chunk_indexes) > 1, "Expected multiple chunks"

            messages = list(reader.messages("/unordered", in_log_time_order=in_log_time_order))
            logging.info(f'pybag: {[message.log_time for message in messages]}')
            logging.info(f'pybag: {[msg.data.data for msg in messages]}')

            expected_log_times = list(range(8)) if in_log_time_order else list(reversed(range(8)))
            assert [msg.log_time for msg in messages] == expected_log_times
            assert [msg.data.data for msg in messages] == [f"msg_{t}" for t in expected_log_times]


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_random_ordered_messages(chunk_size, in_log_time_order: bool, enable_crc_check: bool, profile):
    random.seed(42)  # Make tests reproducible

    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "overlapping.mcap"

        # Create timestamps that will cause overlapping chunks
        sorted_timestamps = list(range(10))
        shuffled_timestamps = random.sample(sorted_timestamps, len(sorted_timestamps))
        logging.info(f'Shuffled timestamps: {shuffled_timestamps}')

        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            for time in shuffled_timestamps:
                writer.write_message("/overlapping", time, String(data=f"msg_{time}"))

        # Read messages with official mcap library
        with open(path, 'rb') as f:
            decoder_factory = Ros2DecoderFactory() if profile == 'ros2' else Ros1DecoderFactory()
            reader = make_reader(f, decoder_factories=[decoder_factory])
            official_mcap_messages = list(reader.iter_decoded_messages(log_time_order=True))
            logging.info(f'mcap: {[msg[-2].log_time for msg in official_mcap_messages]}')

        # Read messages with pybag
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            # Check that we have multiple chunks if configured to do so
            chunk_indexes = reader._reader.get_chunk_indexes()
            logging.info(f'Number of chunks: {len(chunk_indexes)}')
            assert chunk_size is None or len(chunk_indexes) > 1, "Expected multiple chunks"

            messages = list(reader.messages("/overlapping", in_log_time_order=in_log_time_order))
            logging.info(f'pybag: {[msg.log_time for msg in messages]}')
            logging.info(f'pybag: {[msg.data.data for msg in messages]}')

            expected_log_times = sorted_timestamps if in_log_time_order else shuffled_timestamps
            assert [msg.log_time for msg in messages] == expected_log_times
            assert [msg.data.data for msg in messages] == [f"msg_{t}" for t in expected_log_times]


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_duplicate_timestamps(chunk_size, in_log_time_order: bool, enable_crc_check: bool, profile):
    """Test that multiple messages with the same log time are all returned."""
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "duplicate_timestamps.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            # Ensure we have messages split across different chunks if enabled
            timestamp = 1000
            writer.write_message("/test", timestamp, String(data="msg_0"))
            writer.write_message("/test", timestamp, String(data="msg_1"))
            writer.write_message("/test", timestamp, String(data="msg_2"))

        # Read messages with official mcap library
        with open(path, 'rb') as f:
            decoder_factory = Ros2DecoderFactory() if profile == 'ros2' else Ros1DecoderFactory()
            reader = make_reader(f, decoder_factories=[decoder_factory])
            official_mcap_messages = list(reader.iter_decoded_messages(log_time_order=True))
            logging.info(f'mcap: {[msg[-2].log_time for msg in official_mcap_messages]}')

        # Read messages with pybag
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            # Check that we have multiple chunks if configured to do so
            chunk_indexes = reader._reader.get_chunk_indexes()
            logging.info(f'Number of chunks: {len(chunk_indexes)}')
            assert chunk_size is None or len(chunk_indexes) > 1, "Expected multiple chunks"

            messages = list(reader.messages("/test", in_log_time_order=in_log_time_order))
            logging.info(f'pybag: {[msg.log_time for msg in messages]}')
            logging.info(f'pybag: {[msg.data.data for msg in messages]}')

            assert [msg.log_time for msg in messages] == [timestamp] * 3
            for i, message in enumerate(messages):
                assert message.data.data == f"msg_{i}"


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_multi_topic_out_of_order(chunk_size, in_log_time_order: bool, enable_crc_check: bool, profile):
    ahead_messages = [(10, "ahead_0"), (20, "ahead_1"), (30, "ahead_2")]
    behind_messages = [(5, "behind_0"), (15, "behind_1"), (25, "behind_2")]
    expected_per_topic = {"/ahead": ahead_messages, "/behind": behind_messages}

    # Write messages to a bag file
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "multi_topic_pybag.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            for log_time, data in ahead_messages:
                writer.write_message("/ahead", log_time, String(data=data))

            for log_time, data in behind_messages:
                writer.write_message("/behind", log_time, String(data=data))

        # Read messages with official mcap library
        with open(path, 'rb') as f:
            decoder_factory = Ros2DecoderFactory() if profile == 'ros2' else Ros1DecoderFactory()
            reader = make_reader(f, decoder_factories=[decoder_factory])
            for topic in expected_per_topic:
                official_mcap_messages = list(reader.iter_decoded_messages([topic], log_time_order=True))
                logging.info(f'mcap: {[msg[-2].log_time for msg in official_mcap_messages]}')
                logging.info(f'mcap: {[msg[-1].data for msg in official_mcap_messages]}')

        # Read each topic from the bag file
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            for topic, expected in expected_per_topic.items():
                messages = list(reader.messages(topic, in_log_time_order=in_log_time_order))
                assert [msg.log_time for msg in messages] == [log_time for log_time, _ in expected]
                assert [msg.data.data for msg in messages] == [data for _, data in expected]

######################
# Multi MCAP Reader  #
######################

@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_read_multiple_files_as_one(enable_crc_check: bool, profile) -> None:
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        file1 = temp_path / "one.mcap"
        file2 = str(temp_path / "two.mcap")  # str is also Iterable

        with McapFileWriter.open(file1, chunk_size=1, profile=profile) as writer:
            writer.write_message("/chatter", 1, String(data="hello"))
            writer.write_message("/chatter", 3, String(data="again"))
        with McapFileWriter.open(file2, chunk_size=1, profile=profile) as writer:
            writer.write_message("/chatter", 2, String(data="world"))
            writer.write_message("/chatter", 4, String(data="!!"))

        reader = McapMultipleFileReader.from_files([file1, file2], enable_crc_check=enable_crc_check)
        assert reader.start_time == 1 and reader.end_time == 4
        assert reader.get_message_count("/chatter") == 4

        messages = list(reader.messages("/chatter"))
        assert [m.data.data for m in messages] == ["hello", "world", "again", "!!"]

#############################################
# Compatibility with Official MCAP Library  #
#############################################

@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
def test_random_ordered_messages_from_official_mcap(
    chunk_size,
    in_log_time_order: bool,
    enable_crc_check: bool,
):
    random.seed(42)  # Make tests reproducible

    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "overlapping.mcap"

        # Create timestamps that will cause overlapping chunks
        sorted_timestamps = list(range(10))
        shuffled_timestamps = random.sample(sorted_timestamps, len(sorted_timestamps))
        logging.info(f"Shuffled timestamps: {shuffled_timestamps}")

        # Write messages with shuffled timestamps using official mcap library
        with open(path, 'wb') as f:
            writer = Ros2McapWriter(f) if chunk_size is None else Ros2McapWriter(f, chunk_size=chunk_size)
            try:
                schema_id = writer.register_msgdef('std_msgs/String', 'string data\n')
                for timestamp in shuffled_timestamps:
                    writer.write_message(
                        topic='/overlapping',
                        schema=schema_id,
                        message={'data': f'msg_{timestamp}'},
                        log_time=timestamp,
                        publish_time=timestamp,
                    )
            finally:
                writer.finish()

        # Read messages with official mcap library
        with open(path, 'rb') as f:
            reader = make_reader(f, decoder_factories=[Ros2DecoderFactory()])
            official_mcap_messages = list(reader.iter_decoded_messages(log_time_order=True))
            logging.info(f'mcap: {[msg[-2].log_time for msg in official_mcap_messages]}')
            logging.info(f'mcap: {[msg[-1].data for msg in official_mcap_messages]}')

        # Read messages with pybag
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            # Check that we have multiple chunks if configured to do so
            chunk_indexes = reader._reader.get_chunk_indexes()
            logging.info(f'Number of chunks: {len(chunk_indexes)}')
            assert chunk_size is None or len(chunk_indexes) > 1, "Expected multiple chunks"

            messages = list(reader.messages("/overlapping", in_log_time_order=in_log_time_order))
            logging.info(f'pybag: {[msg.log_time for msg in messages]}')
            logging.info(f'pybag: {[msg.data.data for msg in messages]}')

            expected_log_times = sorted_timestamps if in_log_time_order else shuffled_timestamps
            assert [msg.log_time for msg in messages] == expected_log_times
            assert [msg.data.data for msg in messages] == [f"msg_{t}" for t in expected_log_times]


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(1, id="tiny_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
def test_multi_topic_out_of_order_from_official_mcap(
    chunk_size,
    in_log_time_order: bool,
    enable_crc_check: bool,
):
    ahead_messages = [(10, "ahead_0"), (20, "ahead_1"), (30, "ahead_2")]
    behind_messages = [(5, "behind_0"), (15, "behind_1"), (25, "behind_2")]
    expected_per_topic = {"/ahead": ahead_messages, "/behind": behind_messages}

    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "multi_topic_official.mcap"
        with open(path, "wb") as handle:
            writer = Ros2McapWriter(handle) if chunk_size is None else Ros2McapWriter(handle, chunk_size=chunk_size)
            try:
                schema = writer.register_msgdef("std_msgs/String", "string data\n")

                for log_time, data in ahead_messages:
                    writer.write_message(
                        topic="/ahead",
                        schema=schema,
                        message={"data": data},
                        log_time=log_time,
                        publish_time=log_time,
                    )

                for log_time, data in behind_messages:
                    writer.write_message(
                        topic="/behind",
                        schema=schema,
                        message={"data": data},
                        log_time=log_time,
                        publish_time=log_time,
                    )
            finally:
                writer.finish()

        # Read messages with official mcap library
        with open(path, 'rb') as f:
            reader = make_reader(f, decoder_factories=[Ros2DecoderFactory()])
            for topic in expected_per_topic:
                official_mcap_messages = list(reader.iter_decoded_messages([topic], log_time_order=True))
                logging.info(f'mcap: {[msg[-2].log_time for msg in official_mcap_messages]}')
                logging.info(f'mcap: {[msg[-1].data for msg in official_mcap_messages]}')

        # Read each topic from the bag file
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            for topic, expected in expected_per_topic.items():
                messages = list(reader.messages(topic, in_log_time_order=in_log_time_order))
                assert [msg.log_time for msg in messages] == [log_time for log_time, _ in expected]
                assert [msg.data.data for msg in messages] == [data for _, data in expected]

#########################
#  Reverse Iteration    #
#########################

@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_reverse_iteration_multiple_topics(chunk_size, in_log_time_order: bool, enable_crc_check: bool, profile):
    """Test reverse iteration with multiple topics interleaved correctly."""
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "reverse_multi.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            # Write in specific interleaved order
            writer.write_message("/topic1", 10, String(data="t1_10"))
            writer.write_message("/topic2", 5, String(data="t2_5"))
            writer.write_message("/topic1", 3, String(data="t1_3"))
            writer.write_message("/topic2", 15, String(data="t2_15"))

        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            # Reverse iteration
            reverse_messages = list(reader.messages(["/topic1", "/topic2"], in_log_time_order=in_log_time_order, in_reverse=True))
            if in_log_time_order:
                assert [msg.log_time for msg in reverse_messages] == [15, 10, 5, 3]
                assert [msg.data.data for msg in reverse_messages] == ["t2_15", "t1_10", "t2_5", "t1_3"]
            else:
                assert [msg.log_time for msg in reverse_messages] == [15, 3, 5, 10]
                assert [msg.data.data for msg in reverse_messages] == ["t2_15", "t1_3", "t2_5", "t1_10"]

            # Reverse iteration (one topic)
            reverse_messages = list(reader.messages("/topic1", in_log_time_order=in_log_time_order, in_reverse=True))
            if in_log_time_order:
                assert [msg.log_time for msg in reverse_messages] == [10, 3]
                assert [msg.data.data for msg in reverse_messages] == ["t1_10", "t1_3"]
            else:
                assert [msg.log_time for msg in reverse_messages] == [3, 10]
                assert [msg.data.data for msg in reverse_messages] == ["t1_3", "t1_10"]


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("in_log_time_order", [True, False])
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_reverse_iteration_with_time_filter(chunk_size, in_log_time_order: bool, enable_crc_check: bool, profile):
    """Test reverse iteration respects start_time and end_time filters."""
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "reverse_filter.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            for i in range(10):
                writer.write_message("/test", i * 10, String(data=f"msg_{i}"))

        # Reverse iteration with time range [20, 60]
        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            messages = list(reader.messages("/test", start_time=20, end_time=60, in_log_time_order=in_log_time_order, in_reverse=True))
            expected_times = [60, 50, 40, 30, 20]
            assert [msg.log_time for msg in messages] == expected_times


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_reverse_iteration_duplicate_timestamps(chunk_size, enable_crc_check: bool, profile):
    """Test reverse iteration handles duplicate timestamps correctly."""
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "reverse_dup.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            timestamp = 1000
            writer.write_message("/test", timestamp, String(data="msg_0"))
            writer.write_message("/test", timestamp, String(data="msg_1"))
            writer.write_message("/test", timestamp, String(data="msg_2"))

        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            messages = list(reader.messages("/test", in_reverse=True))
            assert [message.data.data for message in messages] == ["msg_2", "msg_1", "msg_0"]


@pytest.mark.parametrize(
    "chunk_size",
    [
        pytest.param(None, id="without_chunks"),
        pytest.param(64, id="with_chunks"),
    ],
)
@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_reverse_iteration_with_filter(chunk_size, enable_crc_check: bool, profile):
    """Test reverse iteration works with message filter callback."""
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "reverse_filter_cb.mcap"
        with McapFileWriter.open(path, chunk_size=chunk_size, chunk_compression=None, profile=profile) as writer:
            for i in range(10):
                writer.write_message("/test", i, String(data=f"msg_{i}"))

        with McapFileReader.from_file(path, enable_crc_check=enable_crc_check) as reader:
            messages = list(reader.messages(
                "/test",
                filter=lambda msg: msg.log_time % 2 == 0,
                in_reverse=True
            ))
            expected_times = [8, 6, 4, 2, 0]
            assert [msg.log_time for msg in messages] == expected_times


@pytest.mark.parametrize("enable_crc_check", [True, False])
@pytest.mark.parametrize("profile", ["ros1", "ros2"])
def test_reverse_iteration_multiple_files(enable_crc_check: bool, profile):
    """Test reverse iteration across multiple MCAP files."""
    String = ros1_std_msgs.String if profile == 'ros1' else ros2_std_msgs.String
    with TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        file1 = temp_path / "one.mcap"
        file2 = temp_path / "two.mcap"

        with McapFileWriter.open(file1, chunk_size=1, profile=profile) as writer:
            writer.write_message("/chatter", 1, String(data="hello"))
            writer.write_message("/chatter", 3, String(data="again"))
        with McapFileWriter.open(file2, chunk_size=1, profile=profile) as writer:
            writer.write_message("/chatter", 2, String(data="world"))
            writer.write_message("/chatter", 4, String(data="!!"))

        reader = McapMultipleFileReader.from_files([file1, file2], enable_crc_check=enable_crc_check)

        # Forward iteration
        forward_messages = list(reader.messages("/chatter", in_reverse=False))
        assert [m.data.data for m in forward_messages] == ["hello", "world", "again", "!!"]
        assert [m.log_time for m in forward_messages] == [1, 2, 3, 4]

        # Reverse iteration
        reverse_messages = list(reader.messages("/chatter", in_reverse=True))
        assert [m.data.data for m in reverse_messages] == ["!!", "again", "world", "hello"]
        assert [m.log_time for m in reverse_messages] == [4, 3, 2, 1]


####################################
#  ROS 2 Char Type Roundtrip Test  #
####################################

def test_mcap_ros2_char_roundtrip():
    """Test that ROS 2 char type is correctly written and read as a string."""
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "char_test.mcap"

        # Write with pybag
        with McapFileWriter.open(path, profile="ros2") as writer:
            msg = ros2_std_msgs.Char(data='A')
            writer.write_message("/char", 1000, msg)

        # Read with pybag
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/char"))
            assert len(messages) == 1
            assert messages[0].data.data == 'A'  # Should be a string, not an integer


def test_mcap_ros2_char_multiple_values():
    """Test that ROS 2 char type handles various character values."""
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "char_multi_test.mcap"

        # Write with pybag - test various characters
        with McapFileWriter.open(path, profile="ros2") as writer:
            writer.write_message("/char", 1000, ros2_std_msgs.Char(data='A'))
            writer.write_message("/char", 2000, ros2_std_msgs.Char(data='z'))
            writer.write_message("/char", 3000, ros2_std_msgs.Char(data='0'))
            writer.write_message("/char", 4000, ros2_std_msgs.Char(data=' '))

        # Read with pybag
        with McapFileReader.from_file(path) as reader:
            messages = list(reader.messages("/char"))
            assert len(messages) == 4
            assert messages[0].data.data == 'A'
            assert messages[1].data.data == 'z'
            assert messages[2].data.data == '0'
            assert messages[3].data.data == ' '


#########################################
#  ROS 1 Types with ROS 2 Profile Tests #
#########################################


@dataclass(kw_only=True)
class Ros1TimeMessage:
    """A message with ROS 1 time type (not valid for ROS 2 profile)."""
    __msg_name__ = 'test_msgs/Ros1TimeMessage'
    stamp: t.ros1.time


@dataclass(kw_only=True)
class Ros1DurationMessage:
    """A message with ROS 1 duration type (not valid for ROS 2 profile)."""
    __msg_name__ = 'test_msgs/Ros1DurationMessage'
    elapsed: t.ros1.duration


def test_mcap_writer_ros2_profile_rejects_ros1_time_type():
    """Test that MCAP writer with ros2 profile rejects ROS 1 time type with helpful error."""
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "ros1_time.mcap"
        with pytest.raises(Ros2MsgError, match="ROS 1 'time' type cannot be used with ROS 2/MCAP"):
            with McapFileWriter.open(path, profile="ros2") as writer:
                msg = Ros1TimeMessage(stamp=t.ros1.Time(secs=1234567890, nsecs=123456789))
                writer.write_message("/time", 1000, msg)


def test_mcap_writer_ros2_profile_rejects_ros1_duration_type():
    """Test that MCAP writer with ros2 profile rejects ROS 1 duration type with helpful error."""
    with TemporaryDirectory() as temp_dir:
        path = Path(temp_dir) / "ros1_duration.mcap"
        with pytest.raises(Ros2MsgError, match="ROS 1 'duration' type cannot be used with ROS 2/MCAP"):
            with McapFileWriter.open(path, profile="ros2") as writer:
                msg = Ros1DurationMessage(elapsed=t.ros1.Duration(secs=100, nsecs=500000000))
                writer.write_message("/duration", 1000, msg)



"""Tests for ROS 1 message MD5 sum computation.

Known MD5 sums from official ROS 1 can be verified with `rosmsg md5 <type>`
"""

import hashlib
from dataclasses import dataclass
from typing import Literal

import pytest

import pybag.types as t
from pybag.schema.ros1msg import Ros1MsgSchemaEncoder, compute_md5sum


class TestComputeMD5Sum:
    """Test the compute_md5sum function."""

    def test_simple_message(self):
        """Test MD5 for a simple message with only builtins."""
        # Message like std_msgs/Int32: int32 data
        msg_def = "int32 data"
        md5 = compute_md5sum(msg_def, "std_msgs/Int32")
        expected_md5 = "da5909fbe378aeaf85e547e830cc1bb7"
        assert md5 == expected_md5

    def test_string_message(self):
        """Test MD5 for std_msgs/String."""
        msg_def = "string data"
        md5 = compute_md5sum(msg_def, "std_msgs/String")
        expected_md5 = "992ce8a1687cec8c8bd883ec73ca41d1"
        assert md5 == expected_md5

    def test_multifield_message(self):
        """Test MD5 for message with multiple fields."""
        # Like geometry_msgs/Vector3
        msg_def = "float64 x\nfloat64 y\nfloat64 z"
        md5 = compute_md5sum(msg_def, "geometry_msgs/Vector3")
        expected_md5 = "4a842b65f413084dc2b10fb484ea7f17"
        assert md5 == expected_md5

    def test_header_message(self):
        """Test MD5 for std_msgs/Header."""
        msg_def = "uint32 seq\ntime stamp\nstring frame_id"
        md5 = compute_md5sum(msg_def, "std_msgs/Header")
        expected_md5 = "2176decaecbce78abc3b96ef049fabed"
        assert md5 == expected_md5

    def test_message_with_constant(self):
        """Test MD5 for message with constants.

        Constants should appear first in the MD5 text.

        Example: a hypothetical message
            uint8 TYPE_A=1
            uint8 TYPE_B=2
            uint8 type
        """
        # Order in msg file: constant, constant, field
        msg_def = "uint8 type\nuint8 TYPE_A=1\nuint8 TYPE_B=2"

        # According to ROS 1 algorithm, constants come first in original order
        # Expected MD5 text: "uint8 TYPE_A=1\nuint8 TYPE_B=2\nuint8 type"
        expected_text = "uint8 TYPE_A=1\nuint8 TYPE_B=2\nuint8 type"
        expected_md5 = hashlib.md5(expected_text.encode()).hexdigest()

        md5 = compute_md5sum(msg_def, "test_msgs/ConstantMsg")
        assert md5 == expected_md5

    def test_message_with_nested_type(self):
        """Test MD5 for message with a nested type.

        For nested types, the MD5 of the nested message should replace the type name.
        """
        # Header MD5 = 2176decaecbce78abc3b96ef049fabed
        # For a message that includes Header:
        # The MD5 text should be: "2176decaecbce78abc3b96ef049fabed header"
        header_md5 = "2176decaecbce78abc3b96ef049fabed"
        expected_text = f"{header_md5} header"
        expected_md5 = hashlib.md5(expected_text.encode()).hexdigest()

        msg_def = f"""std_msgs/Header header
================================================================================
MSG: std_msgs/Header
uint32 seq
time stamp
string frame_id"""
        md5 = compute_md5sum(msg_def, "test_msgs/WithHeader")
        assert md5 == expected_md5

    def test_message_with_array(self):
        """Test MD5 for message with array fields.

        Array notation should be preserved in the MD5 text.
        """
        # A message with dynamic array
        msg_def = "float64[] data"
        expected_text = "float64[] data"
        expected_md5 = hashlib.md5(expected_text.encode()).hexdigest()

        md5 = compute_md5sum(msg_def, "test_msgs/Float64Array")
        assert md5 == expected_md5

    def test_message_with_fixed_array(self):
        """Test MD5 for message with fixed-size array fields."""
        msg_def = "float64[3] position"
        expected_text = "float64[3] position"
        expected_md5 = hashlib.md5(expected_text.encode()).hexdigest()

        md5 = compute_md5sum(msg_def, "test_msgs/FixedArray")
        assert md5 == expected_md5

    def test_comments_stripped(self):
        """Test that comments are stripped from MD5 text."""
        msg_def = "int32 data  # this is a comment"
        # Comment should be stripped
        expected_text = "int32 data"
        expected_md5 = hashlib.md5(expected_text.encode()).hexdigest()

        md5 = compute_md5sum(msg_def, "test_msgs/WithComment")
        assert md5 == expected_md5


class TestSchemaEncoderMD5:
    """Test MD5 computation through the schema encoder."""

    def test_simple_dataclass_md5(self):
        """Test MD5 for a dataclass message matches expected."""
        @dataclass(kw_only=True)
        class SimpleInt:
            __msg_name__ = 'test_msgs/SimpleInt'
            data: t.int32

        encoder = Ros1MsgSchemaEncoder()
        msg_def = encoder.encode(SimpleInt).decode('utf-8')
        md5 = compute_md5sum(msg_def, SimpleInt.__msg_name__)

        # Should match std_msgs/Int32 MD5
        expected_md5 = "da5909fbe378aeaf85e547e830cc1bb7"
        assert md5 == expected_md5

    def test_vector3_like_dataclass_md5(self):
        """Test MD5 for a Vector3-like dataclass."""
        @dataclass(kw_only=True)
        class Vector3:
            __msg_name__ = 'geometry_msgs/Vector3'
            x: t.float64
            y: t.float64
            z: t.float64

        encoder = Ros1MsgSchemaEncoder()
        msg_def = encoder.encode(Vector3).decode('utf-8')
        md5 = compute_md5sum(msg_def, Vector3.__msg_name__)

        expected_md5 = "4a842b65f413084dc2b10fb484ea7f17"
        assert md5 == expected_md5



"""ROS 1 message schema parsing and encoding.

ROS 1 message definitions are similar to ROS 2 but with some key differences:
- time and duration are primitive types (not messages)
- Header is std_msgs/Header (same package as ROS 2)
- Package names don't include '/msg/' in the middle
"""

import dataclasses
import hashlib
import logging
import re
from dataclasses import fields, is_dataclass
from typing import (
    Annotated,
    Any,
    Literal,
    get_args,
    get_origin,
    get_type_hints
)

from pybag.bag.records import ConnectionRecord
from pybag.io.raw_writer import BytesWriter
from pybag.mcap.records import SchemaRecord
from pybag.schema import (
    PRIMITIVE_TYPE_MAP,
    STRING_TYPE_MAP,
    Array,
    Complex,
    Primitive,
    Schema,
    SchemaConstant,
    SchemaDecoder,
    SchemaEncoder,
    SchemaEntry,
    SchemaField,
    SchemaFieldType,
    Sequence,
    String
)
from pybag.types import Message

logger = logging.getLogger(__name__)


class Ros1MsgError(Exception):
    """Exception raised for errors in ROS 1 message parsing."""
    def __init__(self, message: str):
        super().__init__(message)


class _Ros1SchemaParserMixin:
    """Shared parsing logic for ROS 1 schema decoders.

    This mixin provides common functionality for parsing ROS 1 message
    definitions, used by both Ros1MsgSchemaDecoder (for bag files) and
    Ros1McapSchemaDecoder (for MCAP files).
    """

    _builtin_schemas: dict[str, Schema]

    def _create_builtin_schemas(self) -> dict[str, Schema]:
        """Create schemas for built-in ROS 1 message types."""
        builtin_schemas = {}

        # std_msgs/Header
        builtin_schemas['std_msgs/Header'] = Schema(
            'std_msgs/Header',
            {
                'seq': SchemaField(Primitive('uint32'), None),
                'stamp': SchemaField(Primitive('time'), None),
                'frame_id': SchemaField(String('string'), None),
            }
        )

        return builtin_schemas

    def _remove_inline_comment(self, line: str) -> str:
        """Remove inline comments from a line."""
        in_single = False
        in_double = False
        for i, ch in enumerate(line):
            if ch == "'" and not in_double:
                in_single = not in_single
            elif ch == '"' and not in_single:
                in_double = not in_double
            elif ch == '#' and not in_single and not in_double:
                return line[:i].rstrip()
        return line.strip()

    def _parse_value(self, field_type: SchemaFieldType, raw_value: str) -> Any:
        """Parse a default value for a field."""
        if isinstance(field_type, Primitive):
            return PRIMITIVE_TYPE_MAP.get(field_type.type, int)(raw_value)

        if isinstance(field_type, String):
            return raw_value.strip('"') if raw_value.startswith('"') else raw_value.strip("'")

        raise Ros1MsgError('Default values not supported for this field type')

    def _parse_field_type(self, field_raw_type: str, package_name: str) -> SchemaFieldType:
        """Parse a field type string into a SchemaFieldType."""
        # Handle array types: type[N] or type[]
        if array_match := re.match(r'(.*)\[(.*)\]$', field_raw_type):
            element_raw, length_spec = array_match.groups()
            element_field = self._parse_field_type(element_raw, package_name)

            if length_spec == '':
                return Sequence(element_field)
            return Array(element_field, int(length_spec), is_bounded=False)

        # Handle string types
        if field_raw_type == 'string':
            return String('string')

        # Handle primitive types (including ROS 1 specific time/duration)
        if field_raw_type in PRIMITIVE_TYPE_MAP:
            return Primitive(field_raw_type)
        if field_raw_type in ('time', 'duration'):
            return Primitive(field_raw_type)

        # Handle complex types
        if field_raw_type == 'Header':
            return Complex('std_msgs/Header')
        elif '/' not in field_raw_type:
            return Complex(f'{package_name}/{field_raw_type}')
        return Complex(field_raw_type)

    def _parse_field(self, field: str, package_name: str) -> tuple[str, SchemaEntry]:
        """Parse a single field definition."""
        # Extract the field type, name and optional constant value
        if not (match := re.match(r'(\S+)\s+(\S+)(?:\s+(.+))?$', field)):
            raise Ros1MsgError(f'Invalid field definition: {field}')
        field_raw_type, field_raw_name, raw_default = match.groups()

        if not field_raw_type:
            raise Ros1MsgError('Field type cannot be empty')
        if not field_raw_name:
            raise Ros1MsgError('Field name cannot be empty')

        # Check for constants: TYPE NAME=VALUE
        is_constant = '=' in field_raw_name or (raw_default and raw_default.startswith('='))
        if is_constant:
            if '=' in field_raw_name:
                field_raw_name, raw_default = field_raw_name.split('=', 1)
            else:
                raw_default = raw_default[1:].strip()
            if not raw_default:
                raise Ros1MsgError('Constant value cannot be empty')

        # Parse the field type
        schema_type = self._parse_field_type(field_raw_type, package_name)

        # Parse the default value if it exists
        default_value = None
        if raw_default is not None:
            default_value = self._parse_value(schema_type, raw_default)

        if is_constant:
            # default_value is guaranteed to be not None by the control flow above
            if default_value is None:
                raise Ros1MsgError('Constant value cannot be empty')
            return field_raw_name, SchemaConstant(schema_type, default_value)
        return field_raw_name, SchemaField(schema_type, default_value)

    def _add_missing_builtin_schemas(
        self,
        schema_data: bytes,
        sub_schemas: dict[str, Schema]
    ) -> None:
        """Add any missing built-in schemas that are referenced but not defined."""
        schema_text = schema_data.decode('utf-8')
        for builtin_name, builtin_schema in self._builtin_schemas.items():
            if builtin_name not in sub_schemas and builtin_name in schema_text:
                sub_schemas[builtin_name] = builtin_schema

    def _parse_message_definition(
        self,
        msg_name: str,
        msg_def: str,
        schema_data: bytes,
    ) -> tuple[Schema, dict[str, Schema]]:
        """Parse a ROS 1 message definition text into Schema objects.

        Args:
            msg_name: The full message type name (e.g., 'std_msgs/Header').
            msg_def: The message definition text.
            schema_data: The raw schema data bytes (for builtin schema detection).

        Returns:
            Tuple of (main_schema, sub_schemas).
        """
        package_name = msg_name.split('/')[0]

        # Remove comments and empty lines
        lines = [self._remove_inline_comment(line) for line in msg_def.split('\n')]
        lines = [line for line in lines if line]
        msg = '\n'.join(lines)

        # Split along '=' delimiter (80 '=' characters separates message defs)
        msg_parts = [m.strip() for m in msg.split('=' * 80)]

        msg_schema: dict[str, SchemaEntry] = {}
        # The first message does not have the 'MSG: ' prefix line
        main_fields = [m.strip() for m in msg_parts[0].split('\n') if m.strip()]
        for raw_field in main_fields:
            field_name, field = self._parse_field(raw_field, package_name)
            msg_schema[field_name] = field

        sub_msg_schemas: dict[str, Schema] = {}
        for sub_msg in msg_parts[1:]:
            if not sub_msg:
                continue
            # Format: MSG: package/MessageType
            first_line = sub_msg.split('\n')[0].strip()
            if not first_line.startswith('MSG: '):
                continue
            sub_msg_name = first_line[5:]  # Remove 'MSG: ' prefix
            sub_msg_fields = [m.strip() for m in sub_msg.split('\n')[1:] if m.strip()]
            sub_msg_schema: dict[str, SchemaEntry] = {}
            sub_msg_package_name = sub_msg_name.split('/')[0]
            for raw_field in sub_msg_fields:
                field_name, field = self._parse_field(raw_field, sub_msg_package_name)
                sub_msg_schema[field_name] = field
            sub_msg_schemas[sub_msg_name] = Schema(sub_msg_name, sub_msg_schema)

        # Add any required built-in schemas
        main_schema = Schema(msg_name, msg_schema)
        self._add_missing_builtin_schemas(schema_data, sub_msg_schemas)

        return main_schema, sub_msg_schemas


class Ros1MsgSchemaDecoder(_Ros1SchemaParserMixin, SchemaDecoder):
    """Decoder for ROS 1 message definitions from ConnectionRecord (bag files)."""

    def __init__(self):
        self._cache: dict[int, tuple[Schema, dict[str, Schema]]] = {}
        self._builtin_schemas = self._create_builtin_schemas()

    def parse_schema(self, schema: ConnectionRecord) -> tuple[Schema, dict[str, Schema]]:  # type: ignore[override]
        """Parse a ROS 1 message definition into a Schema.

        Args:
            schema: A ConnectionRecord containing the message definition.

        Returns:
            Tuple of (main_schema, sub_schemas).
        """
        if schema.conn in self._cache:
            return self._cache[schema.conn]

        conn_header = schema.connection_header
        result = self._parse_message_definition(
            conn_header.type,
            conn_header.message_definition,
            conn_header.message_definition.encode('utf-8'),
        )

        self._cache[schema.conn] = result
        return result


class Ros1McapSchemaDecoder(_Ros1SchemaParserMixin, SchemaDecoder):
    """Decoder for ROS 1 message definitions from MCAP SchemaRecord.

    This decoder is specifically designed for parsing ROS 1 message schemas
    stored in MCAP files, where the schema is provided as a SchemaRecord
    rather than a ConnectionRecord.
    """

    def __init__(self):
        self._cache: dict[int, tuple[Schema, dict[str, Schema]]] = {}
        self._builtin_schemas = self._create_builtin_schemas()

    def parse_schema(self, schema: SchemaRecord) -> tuple[Schema, dict[str, Schema]]:  # type: ignore[override]
        """Parse a ROS 1 message definition from MCAP SchemaRecord into a Schema.

        Args:
            schema: A SchemaRecord containing the message definition.

        Returns:
            Tuple of (main_schema, sub_schemas).
        """
        if schema.id in self._cache:
            return self._cache[schema.id]

        if schema.encoding != "ros1msg":
            raise Ros1MsgError(f"Expected ros1msg encoding, got: {schema.encoding}")

        result = self._parse_message_definition(
            schema.name,
            schema.data.decode('utf-8'),
            schema.data,
        )

        self._cache[schema.id] = result
        return result


class Ros1MsgSchemaEncoder(SchemaEncoder):
    """Encoder for ROS 1 message definitions."""

    def __init__(self):
        self._cache = None

    @classmethod
    def encoding(cls) -> str:
        """Return the encoding name."""
        return "ros1msg"

    def _extract_literal_int(self, literal_type: Any) -> int:
        """Extract the value from a Literal type annotation."""
        if hasattr(literal_type, '__origin__') and literal_type.__origin__ is Literal:
            if literal_args := get_args(literal_type):
                return int(literal_args[0])
            else:
                raise Ros1MsgError(f"Unknown literal type: {literal_type}")
        if isinstance(literal_type, str) and literal_type.isdigit():
            return int(literal_type)
        if isinstance(literal_type, int):
            return literal_type
        raise Ros1MsgError(f"Unknown literal type: {literal_type}")

    def _parse_annotation(self, annotation_type: Any) -> SchemaFieldType:
        """Parse a type annotation into a SchemaFieldType."""
        annotation_args = get_args(annotation_type)
        if len(annotation_args) < 2:
            raise Ros1MsgError("Field is not correctly annotated.")

        field_type = annotation_args[-1]
        if field_type[0] in PRIMITIVE_TYPE_MAP:
            return Primitive(field_type[0])
        if field_type[0] in ('time', 'duration'):
            return Primitive(field_type[0])

        if field_type[0] in STRING_TYPE_MAP:
            return String(field_type[0])

        if field_type[0] == 'array':
            sub_type = self._parse_annotation(field_type[1])
            if (length := field_type[2]) is None:
                return Sequence(sub_type)
            actual_length = self._extract_literal_int(length)
            return Array(sub_type, length=actual_length, is_bounded=False)

        if field_type[0] == 'complex':
            return Complex(field_type[1])

        if field_type[0] == 'constant':
            return self._parse_annotation(field_type[1])

        raise Ros1MsgError(f"Unknown field type: {field_type}")

    def _parse_default_value(self, annotation: dataclasses.Field) -> Any:
        """Parse the default value from a dataclass field."""
        if annotation.default is not dataclasses.MISSING:
            return annotation.default
        if annotation.default_factory is not dataclasses.MISSING:
            return annotation.default_factory()
        return None

    def _parse_message(self, message: Message | type[Message]) -> tuple[Schema, dict[str, Schema]]:
        """Parse a message dataclass into a Schema."""
        if not is_dataclass(message):
            raise TypeError("Expected a dataclass instance")

        cls = message if isinstance(message, type) else type(message)
        class_name = cls.__msg_name__

        schema = Schema(class_name, {})
        sub_schemas: dict[str, Schema] = {}

        # Use get_type_hints to resolve string annotations from
        # `from __future__ import annotations`
        type_hints = get_type_hints(cls, include_extras=True)

        for field in fields(cls):
            # Get the resolved type from type_hints instead of field.type
            # This handles `from __future__ import annotations` which causes
            # field.type to be a string instead of the actual type
            field_type_hint = type_hints.get(field.name, field.type)

            if get_origin(field_type_hint) is not Annotated:
                if hasattr(field_type_hint, '__msg_name__'):
                    resolved_field_type = Complex(field_type_hint.__msg_name__)
                    field_default = self._parse_default_value(field)
                    schema.fields[field.name] = SchemaField(resolved_field_type, field_default)
                    sub_schema, sub_sub_schemas = self._parse_message(field_type_hint)
                    sub_schemas[sub_schema.name] = sub_schema
                    sub_schemas.update(sub_sub_schemas)
                    continue
                else:
                    raise Ros1MsgError(f"Field '{field.name}' is not correctly annotated.")

            resolved_field_type = self._parse_annotation(field_type_hint)
            field_default = self._parse_default_value(field)

            if get_args(field_type_hint)[-1][0] == 'constant':
                schema.fields[field.name] = SchemaConstant(resolved_field_type, field_default)
                continue
            schema.fields[field.name] = SchemaField(resolved_field_type, field_default)

            # Handle nested complex types
            if isinstance(resolved_field_type, (Sequence, Array)):
                if isinstance(resolved_field_type.type, Complex):
                    list_type = get_args(field_type_hint)[0]
                    if get_origin(list_type) is list and get_args(list_type):
                        complex_annotation = get_args(list_type)[0]
                        if get_origin(complex_annotation) is Annotated:
                            complex_type = get_args(complex_annotation)[0]
                            sub_schema, sub_sub_schemas = self._parse_message(complex_type)
                            sub_schemas[sub_schema.name] = sub_schema
                            sub_schemas.update(sub_sub_schemas)

            if isinstance(resolved_field_type, Complex):
                complex_type = get_args(field_type_hint)[0]
                sub_schema, sub_sub_schemas = self._parse_message(complex_type)
                sub_schemas[sub_schema.name] = sub_schema
                sub_schemas.update(sub_sub_schemas)

        return schema, sub_schemas

    def _type_str(self, field_type: SchemaFieldType) -> str:
        """Convert a SchemaFieldType to a ROS 1 type string."""
        if isinstance(field_type, Primitive):
            return field_type.type
        if isinstance(field_type, String):
            return 'string'
        if isinstance(field_type, Array):
            sub_type = self._type_str(field_type.type)
            return f'{sub_type}[{field_type.length}]'
        if isinstance(field_type, Sequence):
            return f'{self._type_str(field_type.type)}[]'
        if isinstance(field_type, Complex):
            # Remove '/msg/' for ROS 1 style
            return field_type.type.replace('/msg/', '/')
        raise Ros1MsgError(f'Unknown field type: {field_type}')

    def _value_str(self, value: Any) -> str:
        """Convert a value to its string representation."""
        if isinstance(value, bool):
            return 'true' if value else 'false'
        if isinstance(value, (int, float, str, bytes)):
            return str(value)
        if isinstance(value, list):
            return f'[{", ".join(self._value_str(v) for v in value)}]'
        raise Ros1MsgError(f'Unknown value type: {type(value)}')

    def _encode_constant(self, writer: BytesWriter, field_name: str, field: SchemaConstant) -> None:
        """Encode a constant field."""
        encoded_type = self._type_str(field.type)
        encoded_name = field_name.upper()
        encoded_value = self._value_str(field.value)
        writer.write(f'{encoded_type} {encoded_name}={encoded_value}\n'.encode('utf-8'))

    def _encode_field(self, writer: BytesWriter, field_name: str, field: SchemaField) -> None:
        """Encode a regular field."""
        encoded_type = self._type_str(field.type)
        writer.write(f'{encoded_type} {field_name}\n'.encode('utf-8'))

    def encode(self, schema: Message | type[Message]) -> bytes:
        """Encode a message type into ROS 1 message definition format."""
        parsed_schema, sub_schemas = self._parse_message(schema)

        writer = BytesWriter()
        for field_name, field in parsed_schema.fields.items():
            if isinstance(field, SchemaConstant):
                self._encode_constant(writer, field_name, field)
            elif isinstance(field, SchemaField):
                self._encode_field(writer, field_name, field)

        for sub_schema in sub_schemas.values():
            writer.write(('=' * 80 + '\n').encode('utf-8'))
            writer.write(f'MSG: {sub_schema.name.replace("/msg/", "/")}\n'.encode('utf-8'))
            for field_name, field in sub_schema.fields.items():
                if isinstance(field, SchemaConstant):
                    self._encode_constant(writer, field_name, field)
                elif isinstance(field, SchemaField):
                    self._encode_field(writer, field_name, field)

        return writer.as_bytes()

    def parse_schema(self, schema: Message | type[Message]) -> tuple[Schema, dict[str, Schema]]:
        """Parse a message type into a Schema."""
        return self._parse_message(schema)


def compute_md5sum(message_definition: str, msg_type: str) -> str:
    """Compute the MD5 hash for a ROS 1 message definition.

    The MD5 sum is computed following the ROS 1 algorithm:
    1. Remove comments and normalize whitespace
    2. Constants appear first in original order as "type name=value"
    3. For builtin types: "type name"
    4. For complex types: the MD5 of the nested message replaces the type name

    Args:
        message_definition: The full message definition text (may include
            embedded sub-message definitions separated by 80 '=' characters).
        msg_type: The message type name (e.g., 'std_msgs/Header').

    Returns:
        The 32-character hexadecimal MD5 hash.
    """
    # Parse sub-message definitions from the full message definition
    sub_msg_defs = _parse_sub_message_definitions(message_definition)

    # Get the main message definition (first part before any separator)
    main_def = message_definition.split('=' * 80)[0].strip()

    # Compute MD5 text for the main message
    md5_text = _compute_md5_text(main_def, msg_type, sub_msg_defs)

    return hashlib.md5(md5_text.encode('utf-8')).hexdigest()


def _parse_sub_message_definitions(message_definition: str) -> dict[str, str]:
    """Parse embedded sub-message definitions from a full message definition.

    Sub-messages are separated by 80 '=' characters and start with 'MSG: type'.

    Args:
        message_definition: The full message definition text.

    Returns:
        Dictionary mapping message type to its definition text.
    """
    sub_msgs: dict[str, str] = {}

    # Split on the 80 '=' separator
    parts = message_definition.split('=' * 80)

    for part in parts[1:]:  # Skip the first part (main message)
        part = part.strip()
        if not part:
            continue

        lines = part.split('\n')
        first_line = lines[0].strip()

        if first_line.startswith('MSG: '):
            msg_type = first_line[5:].strip()
            # The rest is the message definition
            msg_def = '\n'.join(lines[1:]).strip()
            sub_msgs[msg_type] = msg_def

    return sub_msgs


# ROS 1 builtin types (including time and duration which are special in ROS 1)
_ROS1_BUILTIN_TYPES = {
    'bool', 'byte', 'char',
    'int8', 'uint8', 'int16', 'uint16', 'int32', 'uint32', 'int64', 'uint64',
    'float32', 'float64',
    'string',
    'time', 'duration',
}


def _is_builtin_type(type_name: str) -> bool:
    """Check if a type is a ROS 1 builtin type."""
    # Strip array notation
    bare_type = re.sub(r'\[.*\]$', '', type_name)
    return bare_type in _ROS1_BUILTIN_TYPES


def _compute_md5_text(
    msg_def: str,
    msg_type: str,
    sub_msg_defs: dict[str, str]
) -> str:
    """Compute the canonical MD5 text for a message definition.

    Args:
        msg_def: The message definition (just fields, no embedded types).
        msg_type: The message type name.
        sub_msg_defs: Dictionary of sub-message type -> definition.

    Returns:
        The canonical text to hash for MD5 computation.
    """
    package = msg_type.split('/')[0] if '/' in msg_type else ''

    constants: list[str] = []
    fields: list[str] = []

    for line in msg_def.split('\n'):
        # Remove comments
        if '#' in line:
            line = line[:line.index('#')]
        line = line.strip()
        if not line:
            continue

        # Parse the line to determine if it's a constant or field
        # Constants have the form: TYPE NAME=VALUE
        if '=' in line:
            # It's a constant
            constants.append(line)
        else:
            # It's a field: TYPE NAME
            parts = line.split()
            if len(parts) >= 2:
                field_type = parts[0]
                field_name = parts[1]

                # Get the bare type (without array notation) for type checking
                bare_type = re.sub(r'\[.*\]$', '', field_type)

                if _is_builtin_type(field_type):
                    # Builtin type: use as-is
                    fields.append(f"{field_type} {field_name}")
                else:
                    # Complex type: compute its MD5 and use that instead
                    # Resolve the type name (add package if not specified)
                    if '/' not in bare_type:
                        if bare_type == 'Header':
                            full_type = 'std_msgs/Header'
                        else:
                            full_type = f"{package}/{bare_type}"
                    else:
                        full_type = bare_type

                    # Get the sub-message definition
                    sub_def = sub_msg_defs.get(full_type, '')
                    if not sub_def and full_type == 'std_msgs/Header':
                        # Built-in Header definition
                        sub_def = "uint32 seq\ntime stamp\nstring frame_id"

                    # Recursively compute MD5 for the sub-message
                    sub_md5 = _compute_md5_text(sub_def, full_type, sub_msg_defs)
                    sub_md5_hash = hashlib.md5(sub_md5.encode('utf-8')).hexdigest()

                    fields.append(f"{sub_md5_hash} {field_name}")

    # Combine: constants first, then fields
    result_lines = constants + fields
    return '\n'.join(result_lines)



"""ROS1 nav_msgs message definitions for noetic.

Auto-generated by scripts/generate_ros1_messages.py
Source: https://github.com/ros/common_msgs/tree/noetic-devel/nav_msgs/msg
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

import pybag.ros1.noetic.geometry_msgs as geometry_msgs
import pybag.ros1.noetic.std_msgs as std_msgs
import pybag.types as t


@dataclass(kw_only=True)
class GridCells:
    __msg_name__ = 'nav_msgs/GridCells'

    header: std_msgs.Header
    cell_width: t.float32
    cell_height: t.float32
    cells: t.Array[geometry_msgs.Point]


@dataclass(kw_only=True)
class MapMetaData:
    __msg_name__ = 'nav_msgs/MapMetaData'

    map_load_time: t.ros1.time
    resolution: t.float32
    width: t.uint32
    height: t.uint32
    origin: geometry_msgs.Pose


@dataclass(kw_only=True)
class Odometry:
    __msg_name__ = 'nav_msgs/Odometry'

    header: std_msgs.Header
    child_frame_id: t.string
    pose: geometry_msgs.PoseWithCovariance
    twist: geometry_msgs.TwistWithCovariance


@dataclass(kw_only=True)
class Path:
    __msg_name__ = 'nav_msgs/Path'

    header: std_msgs.Header
    poses: t.Array[geometry_msgs.PoseStamped]


@dataclass(kw_only=True)
class OccupancyGrid:
    __msg_name__ = 'nav_msgs/OccupancyGrid'

    header: std_msgs.Header
    info: MapMetaData
    data: t.Array[t.int8]
